# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#  Standard Imports
import os
import logging
from typing import Any, Dict
import warnings

# Third Party Imports
import h5py
import zarr  # for n5
import numpy.typing as npt

# Local imports
from .pyramidal_data_source import PyramidalDataSource
from ..metadata_sources.bdv_metadata import BigDataViewerMetadata

# Logger Setup
logger = logging.getLogger(__name__.split(".")[1])

# Suppress UserWarnings from zarr.n5
warnings.filterwarnings("ignore", message="Attribute dataType is a reserved N5 keyword")
warnings.filterwarnings(
    "ignore", message="Attribute blockSize is a reserved N5 keyword"
)
warnings.filterwarnings(
    "ignore", message="Attribute dimensions is a reserved N5 keyword"
)


class BigDataViewerDataSource(PyramidalDataSource):
    """BigDataViewer data source.

    This class is used to write data to a BigDataViewer-compatible file. It
    supports both HDF5 and N5 file formats.
    """

    def __init__(self, file_name: str = None, mode: str = "w") -> None:
        """Initializes the BigDataViewerDataSource.

        Parameters
        ----------
        file_name : str
            The name of the file to write to.
        mode : str
            The mode to open the file in. Must be "w" for write or "r" for read.
        """
        #: np.array: The image.
        self.image = None

        #: list: The views.
        self._views = []

        #: zarr.N5Store: The N5 store.
        self.__store = None

        #: str: The file type.
        self.__file_type = os.path.splitext(os.path.basename(file_name))[-1][1:].lower()

        if self.__file_type not in ["h5", "n5"]:
            error_statement = f"Unknown file type {self.__file_type}."
            logger.error(error_statement)
            raise ValueError(error_statement)

        if self.__file_type == "h5":
            self.setup = self._setup_h5
            self.ds_name = self._h5_ds_name

        elif self.__file_type == "n5":
            self.setup = self._setup_n5
            self.ds_name = self._n5_ds_name

        # self._current_frame = 0
        #: BigDataViewerMetadata: The metadata.
        self.metadata = BigDataViewerMetadata()

        super().__init__(file_name=file_name, mode=mode)

    def get_slice(self, x, y, c, z=0, t=0, p=0, subdiv=0) -> npt.ArrayLike:
        """Get a 3D slice of the dataset for a single c, t, p, subdiv.

        Parameters
        ----------
        x : int or slice
            x indices to grab
        y : int or slice
            y indices to grab
        c : int
            Single channel
        z : int or slice
            z indices to grab
        t : int
            Single timepoint
        p : int
            Single position
        subdiv : int
            Subdivision of the dataset to index along

        Returns
        -------
        npt.ArrayLike
            3D (z, y, x) slice of data set
        """
        setup = self.ds_name(t, c, p).replace("???", str(subdiv))
        return self.image[setup][z, y, x]

    def set_metadata_from_configuration_experiment(
        self, configuration: Dict[str, Any], microscope_name: str = None
    ) -> None:
        """Sets the metadata from according to the microscope configuration.

        Parameters
        ----------
        configuration : Dict[str, Any]
            The configuration experiment.
        microscope_name : str
            The microscope name
        """
        # Set rotation and affine transform information in metadata.
        self.metadata.get_affine_parameters(configuration=configuration)
        return super().set_metadata_from_configuration_experiment(
            configuration, microscope_name
        )

    def write(self, data: npt.ArrayLike, **kw) -> None:
        """Writes 2D image to the data source.

        Parameters
        ----------
        data : npt.ArrayLike
            The data to write.
        kw : dict
            The keyword arguments to write.
        """
        self.mode = "w"

        # find current channel
        c, z, t, pos = self._cztp_indices(self._current_frame, self.metadata.per_stack)

        if not (z or c or t or pos):
            self.setup()

        ds_name = self.ds_name(t, c, pos)
        is_kw = len(kw) > 0
        for i in range(self.subdivisions.shape[0]):
            dx, dy, dz = self.resolutions[i, ...]

            # Down-sample in Z.
            if z % dz == 0:
                dataset_name = ds_name.replace("???", str(i))
                # print(z, dz, dataset_name, self.image[dataset_name].shape,
                #       data[::dx, ::dy].shape)
                zs = min(z // dz, self.shapes[i, 0] - 1)  # TODO: Is this necessary?
                try:
                    # Down-sample in X and Y.
                    self.image[dataset_name][zs, ...] = data[::dy, ::dx].astype(self.dtype)
                    if is_kw and (i == 0):
                        self._views.append(kw)
                except OSError as e:
                    if e.errno == 28:
                        logger.error("No disk space left on device. Closing the file.")
                        self.close()
                        raise Exception("No disk space left on device.")

        self._current_frame += 1

        # Check if this was the last frame to write
        c, z, t, pos = self._cztp_indices(self._current_frame, self.metadata.per_stack)
        if (z == 0) and (c == 0) and ((t >= self.shape_t) or (pos >= self.positions)):
            self.setup(
                self.shape_c * self.positions,
                self.shape_c * (pos + 1),
                create_flag=False,
            )
            self.positions = pos + 1

    def _h5_ds_name(self, t, c, p):
        """Get the HDF5 dataset name for the given timepoint, channel, and position.

        Parameters
        ----------
        t : int
            The timepoint.
        c : int
            The channel.
        p : int
            The position.

        Returns
        -------
        ds_name : str
            The dataset name.
        """
        time_group_name = f"t{t:05}"
        setup_group_name = f"s{(c*self.positions+p):02}"
        ds_name = "/".join([time_group_name, setup_group_name, "???", "cells"])
        return ds_name

    def _n5_ds_name(self, t, c, p):
        """Get the N5 dataset name for the given timepoint, channel, and position.

        Parameters
        ----------
        t : int
            The timepoint.
        c : int
            The channel.
        p : int
            The position.

        Returns
        -------
        ds_name : str
            The dataset name.
        """
        time_group_name = f"timepoint{t}"
        setup_group_name = f"setup{(c*self.positions+p)}"
        ds_name = "/".join([setup_group_name, time_group_name, "s???"])
        return ds_name

    def read(self) -> None:
        """Reads data from the image file."""
        self.mode = "r"
        if self.__file_type == "h5":
            self.image = h5py.File(self.file_name, "r")
        elif self.__file_type == "n5":
            self.image = zarr.N5Store(self.file_name)
        xml_fn = os.path.splitext(self.file_name)[0] + ".xml"
        self.metadata.parse_xml(xml_fn)
        self.get_shape_from_metadata()

    def _setup_h5(self, *args, create_flag=True):
        """Set up the HDF5 file.

        This function creates the file and the datasets to populate.

        Parameters
        ----------
        args : tuple
            The setup start and end.
        create_flag : bool
            Flag to create the file.
        """
        if create_flag:
            self.image = h5py.File(self.file_name, "a")

        setup_start, setup_end = 0, self.shape_c * self.positions
        if len(args) >= 2:
            setup_start, setup_end = args[0], args[1]

        # Create setups
        for i in range(setup_start, setup_end):
            setup_group_name = f"s{i:02}"
            if setup_group_name in self.image:
                del self.image[setup_group_name]
            dataset = self.image.create_dataset(
                f"{setup_group_name}/resolutions",
                data=self.resolutions,
                dtype="float64",
            )
            dataset.attrs["element_size_um"] = [self.dy, self.dx, self.dz]

            dataset = self.image.create_dataset(
                f"{setup_group_name}/subdivisions",
                data=self.subdivisions,
                dtype="int32",
            )
            dataset.attrs["element_size_um"] = [self.dy, self.dx, self.dz]

            # https://github.com/bigdataviewer/bigdataviewer-core/issues/102#issuecomment-2072802080
            self.image[setup_group_name].attrs["dataType"] = self.dtype

        # Create the datasets to populate
        for t in range(self.shape_t):
            time_group_name = f"t{t:05}"
            for i in range(setup_start, setup_end):
                setup_group_name = f"s{i:02}"
                for j in range(self.subdivisions.shape[0]):
                    dataset_name = "/".join(
                        [time_group_name, setup_group_name, f"{j}", "cells"]
                    )
                    if dataset_name in self.image:
                        del self.image[dataset_name]
                    # print(f"Creating {dataset_name} with shape {self.shapes[j,...]}")
                    dataset = self.image.create_dataset(
                        dataset_name,
                        chunks=tuple(self.subdivisions[j, ...][::-1]),
                        shape=self.shapes[j, ...],
                        dtype=self.dtype,
                    )
                    dataset.attrs["element_size_um"] = [
                        self.dy * self.resolutions[j, ...][1],
                        self.dx * self.resolutions[j, ...][0],
                        self.dz * self.resolutions[j, ...][2],
                    ]

    def _setup_n5(self, *args, create_flag=True):
        """Set up the N5 file.

        This function creates the file and the datasets to populate. By default,
        it implements blosc compression. Consequently, the anticipated file
        size, and the actual file size, do not match. This is not the case for HDF5.

        Parameters
        ----------
        args : tuple
            The setup start and end.
        create_flag : bool
            Flag to create the file.

        Note
        ----
            Setups and timepoints are flipped in N5 vs. HDF5, see
            https://github.com/bigdataviewer/bigdataviewer-core/blob/master/BDV%20N5%20format.md

        """
        if create_flag:
            self.__store = zarr.N5Store(self.file_name)
            self.image = zarr.group(store=self.__store, overwrite=True)

        setup_start, setup_end = 0, self.shape_c * self.positions
        if len(args) >= 2:
            setup_start, setup_end = args[0], args[1]

        for i in range(setup_start, setup_end):
            setup_group_name = f"setup{i}"
            setup = self.image.create_group(setup_group_name)
            setup.attrs["downsamplingFactors"] = self.resolutions.tolist()
            setup.attrs["dataType"] = self.dtype
            for t in range(self.shape_t):
                time_group_name = f"timepoint{t}"
                timepoint = setup.create_group(time_group_name)
                for j in range(self.subdivisions.shape[0]):
                    s_group_name = f"s{j}"
                    shape = [int(x) for x in self.shapes[j, ...][::-1]]
                    # chunks = self.subdivisions[j, ...]
                    sx = timepoint.zeros(
                        s_group_name, shape=tuple(shape), chunks=(shape[0], shape[1], 1)
                    )
                    sx.attrs["dataType"] = self.dtype
                    sx.attrs["blockSize"] = [shape[0], shape[1], 1]
                    sx.attrs["dimensions"] = list(shape)
        # print(self.image.tree())

    def close(self) -> None:
        """Close the image file."""
        if self._closed:
            return
        self._check_shape(self._current_frame - 1, self.metadata.per_stack)
        if self.__file_type == "n5":
            self.__store.close()
        else:
            self.image.close()
        if self.mode != "r":
            self.metadata.write_xml(self.file_name, views=self._views)
        self._closed = True



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard library imports
import os
from pathlib import Path
from typing import Union

# Third-party imports

# Local application imports
from navigate.tools.file_functions import save_yaml_file
from navigate.tools.decorators import FeatureList, AcquisitionMode
from navigate.config.config import get_navigate_path
from navigate.plugins.plugin_manager import PluginFileManager, PluginPackageManager


class PluginsModel:
    """Plugins manager in the model side"""

    def __init__(self) -> None:
        """Initialize plugins manager class"""
        #: dict: devices dictionary
        self.devices_dict = {}

        #: dict: plugin acquisition modes dictionary
        self.plugin_acquisition_modes = {}

        #: str: feature lists path
        self.feature_lists_path = os.path.join(get_navigate_path(), "feature_lists")

    def load_plugins(self) -> tuple:
        """Load plugins"""
        if not os.path.exists(self.feature_lists_path):
            os.makedirs(self.feature_lists_path)

        plugins_path = os.path.join(Path(__file__).resolve().parent.parent, "plugins")
        plugins_config_path = os.path.join(
            get_navigate_path(), "config", "plugins_config.yml"
        )
        plugin_file_manager = PluginFileManager(plugins_path, plugins_config_path)
        self.load_plugins_through_manager(plugin_file_manager)
        self.load_plugins_through_manager(PluginPackageManager)
        return self.devices_dict, self.plugin_acquisition_modes

    def load_plugins_through_manager(
        self, plugin_manager: Union[PluginFileManager, PluginPackageManager]
    ) -> None:
        """Load plugins through plugin manager

        Parameters
        -------
        plugin_manager : PluginFileManager or PluginPackageManager
            - PluginFileManager
            - PluginPackageManager

        """
        plugins = plugin_manager.get_plugins()

        for _, plugin_ref in plugins.items():
            # read "plugin_config.yml"
            plugin_config = plugin_manager.load_config(plugin_ref)
            if plugin_config is None:
                continue

            # feature
            plugin_manager.load_features(plugin_ref)

            # feature lists
            plugin_manager.load_feature_lists(plugin_ref, self.register_feature_list)

            # acquisition mode
            acquisition_modes = plugin_config.get("acquisition_modes", [])
            plugin_manager.load_acquisition_modes(
                plugin_ref,
                acquisition_modes,
                self.register_acquisition_mode,
            )

            # load devices
            plugin_manager.load_devices(plugin_ref, self.register_device)

    def register_device(self, device, module):
        """Register device

        Parameters
        ----------
        device : str
            device type name
        module : module
            device_startup_functions module
        """
        if module:
            try:
                device_type_name = module.DEVICE_TYPE_NAME
            except AttributeError:
                print(
                    f"Plugin device: {device} is not set correctly!"
                    "Please make sure the DEVICE_TYPE_NAME is given right!"
                )
                return
            # core devices
            core_devices = [
                "camera",
                "remote_focus",
                "galvo",
                "filter_wheel",
                "stage",
                "zoom",
                "shutter",
                "laser",
            ]
            if device_type_name in core_devices:
                try:
                    supported_device_types = module.SUPPORTED_DEVICE_TYPES
                except AttributeError:
                    print(
                        f"Plugin device: {device} doesn't give any information about supported device types"
                        "Please add line 'SUPPORTED_DEVICE_TYPES = []' in device_startup_functions.py in the plugin!"
                    )
                    return
                if device_type_name not in self.devices_dict:
                    self.devices_dict[device_type_name] = {}
                for supported_device_type in supported_device_types:
                    self.devices_dict[device_type_name][supported_device_type] = {
                        "load_device": module.load_device,
                        "start_device": module.start_device,
                    }
            elif device_type_name == "MMCore":
                for device_name in core_devices:
                    if device_name not in self.devices_dict:
                        self.devices_dict[device_name] = {}
                    self.devices_dict[device_name][device_type_name] = {
                        "load_device": module.load_device,
                        "start_device": module.start_device,
                    }
            else:
                self.devices_dict[device_type_name] = {}
                self.devices_dict[device_type_name]["ref_list"] = module.DEVICE_REF_LIST
                self.devices_dict[device_type_name]["load_device"] = module.load_device
                self.devices_dict[device_type_name][ "start_device"] = module.start_device

    def register_acquisition_mode(self, acquisition_mode_name, module):
        """Register acquisition mode

        Parameters
        ----------
        acquisition_mode_name : str
            The name of an acquisition mode
        module : module
            acquisition_mode module
        """
        if not module:
            return
        acquisition_mode = [
            m for m in dir(module) if isinstance(getattr(module, m), AcquisitionMode)
        ]
        if acquisition_mode:
            self.plugin_acquisition_modes[acquisition_mode_name] = getattr(
                module, acquisition_mode[0]
            )(acquisition_mode_name)

    def register_feature_list(self, plugin_feature_list, module):
        """Register feature list

        Parameters
        ----------
        plugin_feature_list : str
            plugin_feature_list path string
        module : module
            feature list module
        """
        features = [
            f for f in dir(module) if isinstance(getattr(module, f), FeatureList)
        ]
        for feature_name in features:
            feature = getattr(module, feature_name)
            feature_list_name = feature.feature_list_name
            feature_list_file_name = "_".join(feature_list_name.split())
            feature_list_content = {
                "module_name": feature_name,
                "feature_list_name": feature_list_name,
                "filename": plugin_feature_list,
            }
            save_yaml_file(
                self.feature_lists_path,
                feature_list_content,
                f"{feature_list_file_name}.yml",
            )



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import logging
import traceback
from typing import Any, Dict

# Third Party Imports
import nidaqmx
from nidaqmx.errors import DaqError
from nidaqmx.constants import LineGrouping

# Local Imports
from navigate.model.devices.laser.base import LaserBase
from navigate.model.devices.device_types import NIDevice
from navigate.tools.decorators import log_initialization

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


@log_initialization
class NILaser(LaserBase, NIDevice):
    """LaserNI Class

    This class is used to control a laser connected to a National Instruments DAQ.
    """

    def __init__(
        self,
        microscope_name: str,
        device_connection: Any,
        configuration: Dict[str, Any],
        device_id: int,
    ) -> None:
        """Initialize the LaserNI class.

        Parameters
        ----------
        microscope_name : str
            The microscope name.
        device_connection : Any
            The device connection object.
        configuration : Dict[str, Any]
            The device configuration.
        device_id : int
            The laser id.
        """
        super().__init__(microscope_name, device_connection, configuration, device_id)
        analog = configuration["configuration"]["microscopes"][microscope_name][
            "laser"
        ][device_id]["power"]["hardware"].get("type", None)

        digital = configuration["configuration"]["microscopes"][microscope_name][
            "laser"
        ][device_id]["onoff"]["hardware"].get("type", None)

        if analog == "NI" and digital == "NI":
            modulation_type = "mixed"
        elif analog == "NI":
            modulation_type = "analog"
        elif digital == "NI":
            modulation_type = "digital"
        else:
            raise ValueError("Laser modulation type not recognized.")

        #: str: The modulation type of the laser - Analog, Digital, or Mixed.
        self.modulation_type = modulation_type

        #: str: Modulation type of the laser - Analog or Digital.
        self.digital_port_type = None

        #: float: The minimum digital modulation voltage.
        self.laser_min_do = None

        #: float: The maximum digital modulation voltage.
        self.laser_max_do = None

        #: nidaqmx.Task: The laser digital modulation task.
        self.laser_do_task = None

        #: float: The minimum analog modulation voltage.
        self.laser_min_ao = None

        #: float: The maximum analog modulation voltage.
        self.laser_max_ao = None

        #: nidaqmx.Task: The laser analog modulation task.
        self.laser_ao_task = None

        #: float: Current laser intensity.
        self._current_intensity = 0

        # Initialize the laser modulation type.
        if self.modulation_type == "mixed":
            self.initialize_digital_modulation()
            self.initialize_analog_modulation()
            logger.info(f"{str(self)} initialized with mixed modulation.")

        elif self.modulation_type == "analog":
            self.initialize_analog_modulation()
            logger.info(f"{str(self)} initialized with analog modulation.")

        elif self.modulation_type == "digital":
            self.initialize_digital_modulation()
            logger.info(f"{str(self)} initialized with digital modulation.")

    def initialize_analog_modulation(self) -> None:
        """Initialize the analog modulation of the laser."""
        try:
            laser_ao_port = self.device_config["power"]["hardware"]["channel"]

            #: float: The minimum analog modulation voltage.
            self.laser_min_ao = self.device_config["power"]["hardware"]["min"]

            #: float: The maximum analog modulation voltage.
            self.laser_max_ao = self.device_config["power"]["hardware"]["max"]

            #: object: The laser analog modulation task.
            self.laser_ao_task = nidaqmx.Task()
            self.laser_ao_task.ao_channels.add_ao_voltage_chan(
                laser_ao_port, min_val=self.laser_min_ao, max_val=self.laser_max_ao
            )
        except DaqError as e:
            logger.debug(f"{str(self)} error:, {e}, {e.error_type}, {e.error_code}")
            print(f"{str(self)} error:, {e}, {e.error_type}, {e.error_code}")

    def initialize_digital_modulation(self) -> None:
        """Initialize the digital modulation of the laser."""
        try:
            laser_do_port = self.device_config["onoff"]["hardware"]["channel"]

            #: float: The minimum digital modulation voltage.
            self.laser_min_do = self.device_config["onoff"]["hardware"]["min"]

            #: float: The maximum digital modulation voltage.
            self.laser_max_do = self.device_config["onoff"]["hardware"]["max"]

            #: object: The laser analog or digital modulation task.
            self.laser_do_task = nidaqmx.Task()

            if "/ao" in laser_do_port:
                # Perform the digital modulation with an analog output port.
                self.laser_do_task.ao_channels.add_ao_voltage_chan(
                    laser_do_port, min_val=self.laser_min_do, max_val=self.laser_max_do
                )
                self.digital_port_type = "analog"

            else:
                # Digital Modulation via a Digital Port
                self.laser_do_task.do_channels.add_do_chan(
                    laser_do_port, line_grouping=LineGrouping.CHAN_FOR_ALL_LINES
                )
                self.digital_port_type = "digital"
        except (KeyError, DaqError) as e:
            self.laser_do_task = None
            if isinstance(e, DaqError):
                logger.debug(e.error_code)
                logger.debug(e.error_type)
                print(e)
                print(e.error_code)
                print(e.error_type)

    def set_power(self, laser_intensity: float) -> None:
        """Sets the analog laser power.

        Parameters
        ----------
        laser_intensity : float
            The laser intensity.
        """
        if self.laser_ao_task is None:
            return
        try:
            scaled_laser_voltage = (int(laser_intensity) / 100) * self.laser_max_ao
            self.laser_ao_task.write(scaled_laser_voltage, auto_start=True)
            self._current_intensity = laser_intensity
        except DaqError as e:
            logger.exception(e)

    def turn_on(self) -> None:
        """Turns on the laser."""

        # Set analog modulation.
        self.set_power(self._current_intensity)

        # If analog only modulation.
        if self.laser_do_task is None:
            return

        # Set digital modulation. Can be analog or digital output type.
        # E.g., AO port can be analog or digital. DO can only be digital.
        try:
            if self.digital_port_type == "digital":
                self.laser_do_task.write(True, auto_start=True)
            elif self.digital_port_type == "analog":
                self.laser_do_task.write(self.laser_max_do, auto_start=True)
        except DaqError as e:
            logger.exception(e)

    def turn_off(self) -> None:
        """Turns off the laser."""
        # set ao power to zero
        tmp = self._current_intensity
        self.set_power(0)
        self._current_intensity = tmp

        if self.laser_do_task is None:
            return
        try:
            if self.digital_port_type == "digital":
                self.laser_do_task.write(False, auto_start=True)
            elif self.digital_port_type == "analog":
                self.laser_do_task.write(self.laser_min_do, auto_start=True)
        except DaqError as e:
            logger.exception(e)

    def close(self) -> None:
        """Close the NI Task before exit."""
        try:
            if self.laser_ao_task is not None:
                self.laser_ao_task.close()
            if self.laser_do_task is not None:
                self.laser_do_task.close()
        except DaqError as e:
            logger.exception(e)

    def __del__(self):
        """Delete the NI Task before exit."""
        if self.laser_ao_task:
            try:
                self.laser_ao_task.close()
            except Exception as e:
                logger.exception(f"Error stopping task: {traceback.format_exc()}")

        if self.laser_do_task:
            try:
                self.laser_do_task.close()
            except Exception as e:
                logger.exception(f"Error stopping task: {traceback.format_exc()}")



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard library imports
import logging
from functools import reduce
import time

# Third party imports

# Local application imports

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class ChangeResolution:
    """
    ChangeResolution class for modifying the resolution mode of a microscope.

    This class provides functionality to change the resolution mode of a microscope by
    reconfiguring the microscope settings and updating the active microscope.

    Notes:
    ------
    - This class is used to change the resolution mode of a microscope by updating the
    microscope settings and configuring the active microscope accordingly.

    - The `resolution_mode` parameter specifies the desired resolution mode, and the
    `zoom_value` parameter specifies the zoom value to be set. These parameters can
    be adjusted to modify the microscope's configuration.

    - The `ChangeResolution` class is typically used to adapt the microscope's settings
    for different imaging requirements during microscopy experiments.

    - The resolution change process involves reconfiguring the microscope, updating the
    active microscope instance, and resuming data acquisition.

    - The `config_table` attribute is used to define the configuration for the
    resolution change process, including signal acquisition and cleanup steps.
    """

    def __init__(self, model, resolution_mode="high", zoom_value="N/A"):
        """Initialize the ChangeResolution class.


        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for resolution mode changes.
        resolution_mode : str, optional
            The desired resolution mode to set for the microscope. Default is "high".
        zoom_value : str, optional
            The zoom value to set for the microscope. Default is "N/A".
        """
        #: MicroscopeModel: The microscope model associated with the resolution change.
        self.model = model

        #: dict: A dictionary defining the configuration for the resolution change
        self.config_table = {
            "signal": {"main": self.signal_func, "cleanup": self.cleanup},
            "node": {"device_related": True},
        }

        #: str: The desired resolution mode to set for the microscope.
        self.resolution_mode = resolution_mode

        #: str: The zoom value to set for the microscope.
        self.zoom_value = zoom_value

    def signal_func(self):
        """Perform actions to change the resolution mode and update the active
         microscope.

        This method carries out actions to change the resolution mode of the microscope
         by reconfiguring the microscope settings, updating the active microscope, and
         resuming data acquisition.

        Returns:
        -------
        bool
            A boolean value indicating the success of the resolution change process.
        """
        # verify microscope name and zoom value
        if (
            self.resolution_mode
            not in self.model.configuration["configuration"]["microscopes"].keys()
        ):
            error_message = f"Can't change resolution: Microscope name {self.resolution_mode} isn't exist!"
            # logger.exception(error_message) doesn't work
            print(error_message)
            raise Exception(error_message)
        if (
            self.zoom_value
            not in self.model.configuration["configuration"]["microscopes"][
                self.resolution_mode
            ]["zoom"]["position"].keys()
        ):
            error_message = (
                f"Can't change resolution: Zoom value {self.zoom_value} isn't exist!"
            )
            # logger.exception(error_message) doesn't work
            print(error_message)
            raise Exception(error_message)
        # check the image size
        camera_config = self.model.configuration["experiment"]["CameraParameters"]
        if (
            camera_config[self.resolution_mode]["img_x_pixels"]
            != camera_config[self.model.active_microscope_name]["img_x_pixels"]
            or camera_config[self.resolution_mode]["img_y_pixels"]
            != camera_config[self.model.active_microscope_name]["img_y_pixels"]
        ):
            error_message = f"Can't change resolution: Image sizes are different!"
            # logger.exception(error_message) doesn't work
            print(error_message)
            raise Exception(error_message)
        # pause data thread
        self.model.pause_data_thread()
        # end active microscope
        self.model.active_microscope.end_acquisition()
        # prepare new microscope
        self.model.configuration["experiment"]["MicroscopeState"][
            "microscope_name"
        ] = self.resolution_mode
        self.model.configuration["experiment"]["MicroscopeState"][
            "zoom"
        ] = self.zoom_value
        self.model.change_resolution(self.resolution_mode)
        logger.debug(f"current resolution is {self.resolution_mode}")
        logger.debug(
            f"current active microscope is {self.model.active_microscope_name}"
        )
        # prepare active microscope
        waveform_dict = self.model.active_microscope.prepare_acquisition()
        self.model.event_queue.put(("waveform", waveform_dict))
        self.model.frame_id = 0
        # prepare channel
        self.model.active_microscope.prepare_next_channel()
        # resume data thread
        self.model.resume_data_thread()
        return True

    def cleanup(self):
        """Perform cleanup actions if needed.

        This method is responsible for performing cleanup actions if required after the
        resolution change process.
        """
        self.model.resume_data_thread()


class SetCameraParameters:
    """
    SetCameraParameters class for modifying the parameters of a camera.

    This class provides functionality to update the parameters of a camera.

    Notes:
    ------
    - This class can set sensor_mode, readout_direction and rolling_shutter_with.

    - If the value of a parameter is None it doesn't update the parameter value.
    """

    def __init__(
        self,
        model,
        microscope_name=None,
        sensor_mode="Normal",
        readout_direction=None,
        rolling_shutter_width=None,
    ):
        """Initialize the ChangeResolution class.


        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for resolution mode changes.
        sensor_mode : str, optional
            The desired sensor mode to set for the camera. "Normal" or "Light-Sheet"
        readout_direction : str, optional
            The readout direction to set for the camera.
            "Top-to-Bottom", "Bottom-to-Top", "Bidirectional" or "Rev. Bidirectional"
        rolling_shutter_width : int, optional
            The number of pixels for the rolling shutter.
        """
        #: MicroscopeModel: The microscope model associated with the resolution change.
        self.model = model

        #: dict: A dictionary defining the configuration for the resolution change
        self.config_table = {
            "signal": {"main": self.signal_func, "cleanup": self.cleanup},
            "node": {"device_related": True},
        }
        #: str: Microscope name
        self.microscope_name = microscope_name

        #: str: The desired sensor mode to set for the camera.
        self.sensor_mode = sensor_mode

        #: str: The reading direction to set for the microscope.
        self.readout_direction = readout_direction

        #: int: The number of pixels for the rolling shutter.
        try:
            self.rolling_shutter_width = int(rolling_shutter_width)
        except (ValueError, TypeError):
            self.rolling_shutter_width = None

    def signal_func(self):
        """Perform actions to change the resolution mode and update the active
         microscope.

        This method carries out actions to change the resolution mode of the microscope
         by reconfiguring the microscope settings, updating the active microscope, and
         resuming data acquisition.

        Returns:
        -------
        bool
            A boolean value indicating the success of the resolution change process.
        """
        if (
            self.microscope_name is None
            or self.microscope_name
            not in self.model.configuration["configuration"]["microscopes"].keys()
        ):
            self.microscope_name = self.model.active_microscope_name
        update_flag = False
        update_sensor_mode = False
        camera_parameters = self.model.configuration["experiment"]["CameraParameters"][
            self.microscope_name
        ]
        camera_config = self.model.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["camera"]
        updated_value = [None] * 4
        updated_value[0] = self.microscope_name
        if (
            self.sensor_mode in ["Normal", "Light-Sheet"]
            and self.sensor_mode != camera_parameters["sensor_mode"]
        ):
            update_flag = True
            update_sensor_mode = True
            camera_parameters["sensor_mode"] = self.sensor_mode
            updated_value[1] = self.sensor_mode
        if camera_parameters["sensor_mode"] == "Light-Sheet":
            if self.readout_direction in camera_config[
                "supported_readout_directions"
            ] and (
                update_sensor_mode
                or camera_parameters["readout_direction"] != self.readout_direction
            ):
                update_flag = True
                camera_parameters["readout_direction"] = self.readout_direction
                updated_value[2] = self.readout_direction
            if self.rolling_shutter_width and (
                update_sensor_mode
                or self.rolling_shutter_width != camera_parameters["number_of_pixels"]
            ):
                update_flag = True
                camera_parameters["number_of_pixels"] = self.rolling_shutter_width
                updated_value[3] = self.rolling_shutter_width

        if not update_flag or self.microscope_name != self.model.active_microscope_name:
            return True
        # pause data thread
        self.model.pause_data_thread()
        # end active microscope
        self.model.active_microscope.end_acquisition()
        # set parameters and prepare active microscope
        waveform_dict = self.model.active_microscope.prepare_acquisition()
        self.model.event_queue.put(("waveform", waveform_dict))
        self.model.event_queue.put(("display_camera_parameters", updated_value))
        self.model.frame_id = 0
        # prepare channel
        self.model.active_microscope.prepare_next_channel()
        # resume data thread
        self.model.resume_data_thread()
        return True

    def cleanup(self):
        self.model.resume_data_thread()


class UpdateExperimentSetting:
    def __init__(self, model, experiment_parameters={}):
        self.model = model

        #: dict: A dictionary defining the configuration for the resolution change
        self.config_table = {
            "signal": {"main": self.signal_func, "cleanup": self.cleanup},
            "node": {"device_related": True},
        }

        self.experiment_parameters = experiment_parameters

    def signal_func(self):
        """Perform actions to change the resolution mode and update the active
         microscope.

        This method carries out actions to change the resolution mode of the microscope
         by reconfiguring the microscope settings, updating the active microscope, and
         resuming data acquisition.

        Returns:
        -------
        bool
            A boolean value indicating the success of the resolution change process.
        """
        if type(self.experiment_parameters) != dict:
            return False
        # pause data thread
        self.model.pause_data_thread()
        # end active microscope
        self.model.active_microscope.end_acquisition()

        state = self.model.configuration["experiment"]["MicroscopeState"]
        pre_z_steps = state["number_z_steps"]
        pre_channels = sum(
            [v["is_selected"] is True for k, v in state["channels"].items()]
        )
        pre_timepoints = state["timepoints"]

        # update experiment values
        # check if any parameter about x, y, c, z, t changed
        for k, v in self.experiment_parameters.items():
            try:
                parameters = k.split(".")
                config_ref = reduce(lambda pre, n: f"{pre}['{n}']", parameters, "")
                exec(f"self.model.configuration['experiment']{config_ref} = {v}")
            except Exception as e:
                logger.error(f"*** parameter {k} failed to update to value {v}")
                logger.error(e)
        # set parameters and prepare active microscope
        waveform_dict = self.model.active_microscope.prepare_acquisition()
        self.model.event_queue.put(("waveform", waveform_dict))
        self.model.frame_id = 0
        # prepare channel
        self.model.active_microscope.prepare_next_channel()
        # update image writer
        if self.model.image_writer:
            z_steps = state["number_z_steps"]
            channels = sum(
                [v["is_selected"] is True for k, v in state["channels"].items()]
            )
            timepoints = state["timepoints"]
            if pre_z_steps != z_steps or pre_channels != channels or pre_timepoints != timepoints:
                self.model.image_writer.initialize_saving(sub_dir=time.strftime("%H%M%S"))
            try:
                self.model.image_writer.data_source.set_metadata_from_configuration_experiment(
                    self.model.configuration
                )
            except Exception as e:
                logger.exception(f"Update image writer metadata failed: {e}")
        # resume data thread
        self.model.resume_data_thread()
        return True

    def cleanup(self):
        self.model.resume_data_thread()



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import logging
import time
import ctypes
from typing import Optional, Any, Dict, List

# Third Party Imports
import numpy as np
from tifffile import TiffFile, TiffFileError

# Local Imports
from navigate.model.analysis import camera
from navigate.model.concurrency.concurrency_tools import SharedNDArray
from navigate.model.devices.camera.base import CameraBase
from navigate.tools.decorators import log_initialization

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


@log_initialization
class SyntheticCameraController:
    """SyntheticCameraController. Synthetic Camera API."""

    def __init__(self) -> None:
        """Initialize SyntheticCameraController class."""
        pass

    @staticmethod
    def get_property_value(_: str) -> Any:
        """Provides the property value after looking it up in the property_dict

        Parameters
        ----------
        _ : str
            Not currently used.

        Returns
        -------
        property_value : int
            Currently hard-coded to -1.
        """
        return -1

    @staticmethod
    def set_property_value(name, value):
        """Set the value of a camera property.

        Parameters
        ----------
        name : str
            Name of the property to set.
        value : int
            Value to set the property to.
        """
        logger.info(f"Camera property {name} set to {value}")


@log_initialization
class SyntheticCamera(CameraBase):
    """SyntheticCamera camera class."""

    def __init__(
        self,
        microscope_name: str,
        device_connection: Any,
        configuration: Dict[str, Any],
        *args: Optional[Any],
        **kwargs: Optional[Any],
    ) -> None:
        """Initialize SyntheticCamera class.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_connection : Any
            Hardware device to connect to
        configuration : Dict[str, Any]
            Global configuration of the microscope
        """
        super().__init__(microscope_name, device_connection, configuration)

        #: str: Name of the microscope
        self.microscope_name = microscope_name

        #: Any: Device connection
        self.device_connection = device_connection

        #: Dict[str, Any]: Configuration settings
        self.configuration = configuration

        #: bool: Whether the camera is currently acquiring
        self.is_acquiring = False

        #: int: mean background count for synthetic image
        self._mean_background_count = 100

        #: float: noise sigma for synthetic image
        self._noise_sigma = camera.compute_noise_sigma()

        #: int: current image id
        self.current_frame_idx = None

        #: object: data buffer
        self.data_buffer = None

        #: int: number of frames
        self.num_of_frame = None

        #: int: previous image id
        self.pre_frame_idx = None

        #: bool: whether to use random image
        self.random_image = True

        #: int: serial number
        self.serial_number = "synthetic"

        #: float: exposure time
        self.camera_exposure_time = 0.2

        #: int: x binning
        self.x_binning = 1

        #: int: y binning
        self.y_binning = 1

        #: int: width
        self.x_pixels = self.camera_parameters["x_pixels"]

        #: int: height
        self.y_pixels = self.camera_parameters["y_pixels"]

        #: int: center x
        self.center_x = self.x_pixels // 2

        #: int: center y
        self.center_y = self.y_pixels // 2

        #: int: current image id
        self.img_id = 0

        #: int: current tif id
        self.current_tif_id = 0

        #: list: list of tif images
        self.tif_images = []

        self.camera_parameters["supported_trigger_sources"] = ["External", "Internal"]

    def __str__(self) -> str:
        """String representation of SyntheticCamera class.

        Returns
        -------
        representation : str
            String representation of SyntheticCamera class.
        """
        return "SyntheticCamera"

    def __del__(self) -> None:
        """Delete SyntheticCamera class."""
        pass

    def report_settings(self) -> None:
        """Print Camera Settings."""
        pass

    def close_camera(self) -> None:
        """Close SyntheticCamera Camera"""
        pass

    def set_sensor_mode(self, mode: str) -> None:
        """Set SyntheticCamera sensor mode.

        Parameters
        ----------
        mode : str
            'Normal' or 'Light-Sheet'
        """
        pass

    def set_exposure_time(self, exposure_time: float) -> None:
        """Set SyntheticCamera exposure time.

        All of our units are in milliseconds. Function converts to seconds.

        Parameters
        ----------
        exposure_time : float
            Exposure time in seconds.
        """
        self.camera_exposure_time = exposure_time

    def set_line_interval(self, line_interval_time: float) -> None:
        """Set SyntheticCamera line interval.

        Parameters
        ----------
        line_interval_time : float
            Line interval duration.
        """
        pass

    def set_binning(self, binning_string: str) -> bool:
        """Set SyntheticCamera binning mode.

        Parameters
        ----------
        binning_string : str
            Desired binning properties (e.g., '2x2', '4x4', '8x8')

        Returns
        -------
        bool
            True if successful, False otherwise.
        """
        binning_dict = {
            "1x1": 1,
            "2x2": 2,
            "4x4": 4,
            # '8x8': 8,
            # '16x16': 16,
            # '1x2': 102,
            # '2x4': 204
        }
        if binning_string not in binning_dict.keys():
            logger.debug(f"can't set binning to {binning_string}")
            print(f"can't set binning to {binning_string}")
            return False
        
        self.x_binning = int(binning_string[0])
        self.y_binning = int(binning_string[2])
        self.x_pixels = int(self.x_pixels / self.x_binning)
        self.y_pixels = int(self.y_pixels / self.y_binning)
        return True

    def initialize_image_series(
        self,
        data_buffer: Optional[List[SharedNDArray]] = None,
        number_of_frames: int = 100,
    ):
        """Initialize SyntheticCamera image series.

        Parameters
        ----------
        data_buffer : Optional[List[SharedNDArray]]
            The shared data buffer. Default is None.
        number_of_frames : int
            Number of frames.  Default is 100.
        """
        self.data_buffer = data_buffer
        self.num_of_frame = number_of_frames
        self.current_frame_idx = 0
        self.pre_frame_idx = 0
        self.is_acquiring = True

    def close_image_series(self) -> None:
        """Close image series.

        Stops the acquisition and sets is_acquiring flag to False.
        """
        self.pre_frame_idx = 0
        self.current_frame_idx = 0
        self.is_acquiring = False

    def load_images(self, filenames: Optional[str] = None, ds=None) -> None:
        """Pre-populate the buffer with images. Can either come from TIFF files or
        Numpy stacks."""
        self.random_image = False
        #: int: current image id
        self.img_id = 0
        #: int: current tif id
        self.current_tif_id = 0
        #: list: list of tif images
        self.tif_images = []
        idx = 0
        if filenames is not None:
            # Load TIFF file into buffer as slices
            for image_file in filenames:
                try:
                    tif = TiffFile(image_file)
                    if len(tif.pages) == 1:
                        self.tif_images.append([tif.asarray()])
                    else:
                        self.tif_images.append(tif.asarray())
                    idx += len(tif.pages)
                    if idx >= self.num_of_frame:
                        return
                except TiffFileError:
                    pass
        elif ds is not None:
            # Load a Numpy stack into buffer as slices
            # Assume the stack is in the order ZYX
            for idx, data in enumerate(ds):
                self.tif_images.append(data)
                idx += len(data)
                if idx >= self.num_of_frame:
                    return
        else:
            self.random_image = True

    def generate_new_frame(self) -> None:
        """Generate a synthetic image."""
        if not self.is_acquiring:
            return
        if self.random_image:
            image = np.random.normal(
                0,
                self._noise_sigma
                / 0.47,  # TODO: Don't hardcode 0.47 electrons per count
                size=(self.x_pixels, self.y_pixels),
            ).astype(np.uint16) + int(self._mean_background_count)
        else:
            image = self.tif_images[self.current_tif_id][self.img_id]
            self.img_id += 1
            if self.img_id >= len(self.tif_images[self.current_tif_id]):
                self.img_id = 0
                self.current_tif_id = (self.current_tif_id + 1) % len(self.tif_images)

        ctypes.memmove(
            self.data_buffer[self.current_frame_idx].ctypes.data,
            image.ctypes.data,
            self.x_pixels * self.y_pixels * 2,
        )

        self.current_frame_idx = (self.current_frame_idx + 1) % self.num_of_frame

    def get_new_frame(self) -> List[int]:
        """Get frame from SyntheticCamera camera."""

        time.sleep(self.camera_exposure_time)
        timeout = 500
        while self.pre_frame_idx == self.current_frame_idx and timeout:
            time.sleep(0.001)
            timeout -= 1
        if timeout <= 0:
            return []
        if self.pre_frame_idx < self.current_frame_idx:
            frames = list(range(self.pre_frame_idx, self.current_frame_idx))
        else:
            frames = list(range(self.pre_frame_idx, self.num_of_frame))
            frames += list(range(0, self.current_frame_idx))
        self.pre_frame_idx = self.current_frame_idx
        return frames

    def set_ROI(
        self,
        roi_width: int = 2048,
        roi_height: int = 2048,
        center_x: int = 1024,
        center_y: int = 1024,
    ) -> bool:
        """Change the size of the active region on the camera.

        Parameters
        ----------
        roi_width : int
            Width of active camera region.
        roi_height : int
            Height of active camera region.
        center_x : int
            X position of the center of view
        center_y : int
            Y position of the center of view
        
        Returns
        -------
        bool
            True if successful, False otherwise.
        """
        self.x_pixels = roi_width
        self.y_pixels = roi_height
        self.center_x = center_x
        self.center_y = center_y
        return True

    @staticmethod
    def calculate_readout_time() -> float:
        """Calculate duration of time needed to read out an image.

        Calculates the readout time and maximum frame rate according to the camera
        configuration settings.

        Returns
        -------
        readout_time : float
            Duration of time needed to read out an image.
        """
        readout_time = 0.01  # 10 milliseconds.
        return readout_time

    def set_trigger_mode(self, trigger_source: str="External") -> None:
        """Set camera trigger mode.
        Parameters
        ----------
        trigger_source : str
            Trigger source, either 'External' or 'Internal'.
        """
        logger.debug(f"Set camera trigger mode: {trigger_source}")



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

#  Standard Library Imports
import logging
import time
import traceback
from typing import Any, Dict

# Third Party Imports

# Local Imports
from navigate.model.devices.laser.base import LaserBase
from navigate.model.devices.device_types import SerialDevice
from navigate.model.devices.APIs.asi.asi_tiger_controller import TigerController
from navigate.tools.decorators import log_initialization

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


@log_initialization
class ASILaser(LaserBase, SerialDevice):
    """ASILaser - Class for controlling ASI Lasers

    This class is used to control a laser connected to a ASI Device.
    """

    def __init__(self, microscope_name, device_connection, configuration, device_id: int = 0):
        """Initialize the ASILaser class.

        Parameters
        ----------
        microscope_name : str
            The microscope name.
        device_connection : TigerController
            The device connection object.
        configuration : dict
            The device configuration.
        device_id : int
            The laser id.
        """

        super().__init__(microscope_name, device_connection, configuration, device_id)
        analog = configuration["configuration"]["microscopes"][microscope_name][
            "laser"
        ][device_id]["power"]["hardware"].get("type", None)

        digital = configuration["configuration"]["microscopes"][microscope_name][
            "laser"
        ][device_id]["onoff"]["hardware"].get("type", None)

        if analog == "asi.ASI" and digital == "asi.ASI":
            modulation_type = "mixed"
            #: float: The minimum digital modulation voltage.
            self.laser_min_do = self.device_config["onoff"]["hardware"]["min"]
            #: float: The maximum digital modulation voltage.
            self.laser_max_do = self.device_config["onoff"]["hardware"]["max"]
            #: float: The minimum analog modulation voltage.
            self.laser_min_ao = self.device_config["power"]["hardware"]["min"]
            #: float: The maximum analog modulation voltage.
            self.laser_max_ao = self.device_config["power"]["hardware"]["max"]
            #: str: Output axes on Tiger Controller
            self.analog_axis = self.device_config["power"]["hardware"]["axis"]
            self.digital_axis = self.device_config["onoff"]["hardware"]["axis"]
        elif analog == "asi.ASI":
            modulation_type = "analog"
            #: float: The minimum analog modulation voltage.
            self.laser_min_ao = self.device_config["power"]["hardware"]["min"]
            #: float: The maximum analog modulation voltage.
            self.laser_max_ao = self.device_config["power"]["hardware"]["max"]
            #: str: Output axis on Tiger Controller
            self.analog_axis = self.device_config["power"]["hardware"]["axis"]
        elif digital == "asi.ASI":
            modulation_type = "digital"
            #: float: The minimum digital modulation voltage.
            self.laser_min_do = self.device_config["onoff"]["hardware"]["min"]
            #: float: The maximum digital modulation voltage.
            self.laser_max_do = self.device_config["onoff"]["hardware"]["max"]
            #: str: Output axes on Tiger Controller
            self.digital_axis = self.device_config["onoff"]["hardware"]["axis"]
        else:
            raise ValueError("Laser modulation type not recognized.")

        #: str: The modulation type of the laser - Analog, Digital, or Mixed.
        self.modulation_type = modulation_type

        #: TigerController: ASI Tiger Controller object.
        self.laser = device_connection

        #: float: Current laser intensity.
        self._current_intensity = 0

    def __str__(self):
        """String representation of the class."""
        return "ASILaser"

    @classmethod
    def connect(cls, port, baudrate=115200, timeout=0.25):
        """Build ASILaser Serial Port connection

        Parameters
        ----------
        port : str
            Port for communicating with the filter wheel, e.g., COM1.
        baudrate : int
            Baud rate for communicating with the filter wheel, default is 115200.
        timeout : float
            Timeout for communicating with the filter wheel, default is 0.25.

        Returns
        -------
        tiger_controller : TigerController
            ASI Tiger Controller object.
        """
        # wait until ASI device is ready
        tiger_controller = TigerController(port, baudrate)
        tiger_controller.connect_to_serial()
        if not tiger_controller.is_open():
            logger.error("ASI stage connection failed.")
            raise Exception("ASI stage connection failed.")
        return tiger_controller
    
    def set_power(self, laser_intensity: float):
        """Sets the analog laser power.

        Parameters
        ----------
        laser_intensity : float
            The laser intensity.
        """
        self.output_voltage = (int(laser_intensity) / 100) * self.laser_max_ao * 1000
        if self.output_voltage > (self.laser_max_ao * 1000):
            self.output_voltage = self.laser_max_ao * 1000
        self.laser.move_axis(self.analog_axis, self.output_voltage)
        self._current_intensity = laser_intensity

    def turn_on(self):
        """Turns on the laser."""
        if self.modulation_type == "mixed":
            self.set_power(self._current_intensity)
            self.laser.logic_card_on(self.digital_axis)
            logger.info(f"{str(self)} initialized with mixed modulation.")

        elif self.modulation_type == "analog":
            self.set_power(self._current_intensity)
            logger.info(f"{str(self)} initialized with analog modulation.")

        elif self.modulation_type == "digital":
            self.laser.logic_card_on(self.digital_axis)
            logger.info(f"{str(self)} initialized with digital modulation.")
        

    def turn_off(self):
        """Turns off the laser."""
        if self.modulation_type == "mixed":
            tmp = self._current_intensity
            self.set_power(0)
            self._current_intensity = tmp
            self.laser.logic_card_off(self.digital_axis)
            logger.info(f"{str(self)} initialized with mixed modulation.")

        elif self.modulation_type == "analog":
            tmp = self._current_intensity
            self.set_power(0)
            self._current_intensity = tmp
            logger.info(f"{str(self)} initialized with analog modulation.")

        elif self.modulation_type == "digital":
            self.laser.logic_card_off(self.digital_axis)
            logger.info(f"{str(self)} initialized with digital modulation.")

    
    def close(self):
        """Close the ASI Laser serial port.

        Turns the laser off and then closes the port.
        """
        if self.laser.is_open():
            self.turn_off()
            logger.debug("ASI Laser - Closing Device.")
            self.laser.disconnect_from_serial()

    def __del__(self):
        """Destructor for the ASILaser class."""
        self.close()



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard library imports
import tkinter as tk
from tkinter import ttk
from navigate.view.custom_widgets.popup import PopUp
from navigate.view.custom_widgets.LabelInputWidgetFactory import LabelInput
from navigate.view.custom_widgets.validation import ValidatedSpinbox
import logging

# Logging
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class WaveformParameterPopupWindow:
    """Popup window with waveform parameters for galvos, remote focusing, etc."""

    def __init__(self, root, configuration_controller, *args, **kwargs):
        # Creating popup window with this name and size/placement, PopUp is a
        # Toplevel window
        #: PopUp: The popup window
        self.popup = PopUp(
            root,
            name="Waveform Parameter Settings",
            size="+320+180",
            top=False,
            transient=False
        )

        #: configuration_controller: The configuration controller
        self.configuration_controller = configuration_controller

        # Storing the content frame of the popup, this will be the parent of
        # the widgets
        content_frame = self.popup.get_frame()
        content_frame.columnconfigure(0, pad=5)
        content_frame.columnconfigure(1, pad=5)
        content_frame.rowconfigure(0, pad=5)
        content_frame.rowconfigure(1, pad=5)
        content_frame.rowconfigure(2, pad=5)

        # Formatting
        tk.Grid.columnconfigure(content_frame, "all", weight=1)
        tk.Grid.rowconfigure(content_frame, "all", weight=1)

        #: dict: Dictionary for all the variables
        self.inputs = {}

        #: dict: Dictionary for all the buttons
        self.buttons = {}

        #: ttk.Frame: Frame for mode and magnification
        self.mode_mag_frame = ttk.Frame(content_frame, padding=(0, 0, 0, 0))

        #: ttk.Frame: Frame for saving waveform parameters
        self.save_frame = ttk.Frame(content_frame, padding=(0, 0, 0, 0))

        #: ttk.LabelFrame: Frame for remote focus parameters
        self.remote_focus_frame = ttk.LabelFrame(
            content_frame, text="Remote Focus Settings", padding=(0, 0, 0, 0)
        )

        #: ttk.LabelFrame: Frame for galvo parameters
        self.galvo_frame = ttk.LabelFrame(
            content_frame, text="Galvo Settings", padding=(0, 0, 0, 0)
        )

        #: ttk.LabelFrame: Frame for waveform
        self.waveform_frame = ttk.LabelFrame(
            content_frame, text="Waveform Parameters", padding=(0, 0, 0, 0)
        )

        # Griding Frames
        self.mode_mag_frame.grid(row=0, column=0, sticky=tk.NSEW, pady=(10, 0))
        self.save_frame.grid(row=0, column=3, sticky=tk.NE, padx=10, pady=(10, 0))
        self.remote_focus_frame.grid(
            row=2, column=0, columnspan=4, sticky=tk.NSEW, padx=10, pady=(10, 0)
        )
        self.galvo_frame.grid(
            row=3, column=0, columnspan=4, sticky=tk.NSEW, padx=10, pady=(10, 0)
        )
        self.waveform_frame.grid(
            row=4, column=0, columnspan=4, sticky=tk.NSEW, padx=10, pady=10
        )

        # Filling Frames with widgets
        max_length = max(len("Microscope"), len("Magnification"))
        padded_text = "Microscope"
        padded_text = padded_text.ljust(max_length)
        self.inputs["Mode"] = LabelInput(
            parent=self.mode_mag_frame,
            label=padded_text,
            input_class=ttk.Combobox,
            input_var=tk.StringVar(),
            label_args={"padding": (2, 5, 5, 0)},
        )
        self.inputs["Mode"].grid(row=0, column=0, padx=10)
        self.inputs["Mode"].pad_input(50, 5, 0, 0)
        self.inputs["Mode"].state(["readonly"])

        padded_text = "Magnification"
        padded_text = padded_text.ljust(max_length)
        self.inputs["Mag"] = LabelInput(
            parent=self.mode_mag_frame,
            label=padded_text,
            input_class=ttk.Combobox,
            input_var=tk.StringVar(),
            label_args={"padding": (2, 5, 5, 0)},
        )
        self.inputs["Mag"].grid(row=1, column=0, padx=10)
        self.inputs["Mag"].pad_input(50, 5, 0, 0)
        self.inputs["Mag"].state(["readonly"])

        # Save Waveform Parameters Frame
        self.buttons["Save"] = ttk.Button(self.save_frame, text="Save Configuration")
        self.buttons["Save"].grid(
            row=0, column=0, sticky=tk.NSEW, padx=(5, 0), pady=(0, 0)
        )

        # Toggle Waveform Button
        self.buttons["toggle_waveform_button"] = ttk.Button(
            self.save_frame,
            text="Disable Waveforms",
        )
        self.buttons["toggle_waveform_button"].grid(
            row=1, column=0, sticky=tk.NSEW, padx=(5, 0), pady=(0, 0)
        )

        # Laser Frame
        laser_labels = self.configuration_controller.lasers_info
        title_labels = ["Laser", "Amplitude", "Offset"]
        # Loop for widgets
        for i in range(len(title_labels)):
            # Title labels
            title = ttk.Label(
                self.remote_focus_frame, text=title_labels[i], padding=(2, 5, 60, 0)
            )
            title.grid(row=0, column=i, sticky=tk.NSEW, padx=(0, 5))
        for i, label in enumerate(laser_labels):
            # Laser labels
            laser = ttk.Label(
                self.remote_focus_frame, text=laser_labels[i], padding=(2, 5, 0, 0)
            )
            laser.grid(row=i + 1, column=0, sticky=tk.NSEW)

            # Entry Widgets
            self.inputs[laser_labels[i] + " Amp"] = LabelInput(
                parent=self.remote_focus_frame,
                input_class=ValidatedSpinbox,
                input_var=tk.StringVar(),
            )

            self.inputs[laser_labels[i] + " Amp"].grid(
                row=i + 1, column=1, sticky=tk.NSEW, pady=(20, 0), padx=(0, 5)
            )

            self.inputs[laser_labels[i] + " Off"] = LabelInput(
                parent=self.remote_focus_frame,
                input_class=ValidatedSpinbox,
                input_var=tk.StringVar(),
            )

            self.inputs[laser_labels[i] + " Off"].grid(
                row=i + 1, column=2, sticky=tk.NSEW, pady=(20, 0)
            )

        galvo_labels = list(
            map(lambda i: f"Galvo {i}", range(self.configuration_controller.galvo_num))
        )

        title_labels = ["Galvo", "Amplitude", "Offset", "Rising Ramp(%)", "Frequency"]
        # Loop for widgets
        for i in range(len(title_labels)):
            # Title labels
            title = ttk.Label(
                self.galvo_frame, text=title_labels[i], padding=(2, 5, 53, 0)
            )
            title.grid(row=0, column=i, sticky=tk.NSEW, padx=(0, 5))

        for i, label in enumerate(galvo_labels):
            galvo = ttk.Label(
                self.galvo_frame, text=galvo_labels[i], padding=(2, 5, 53, 0)
            )

            galvo.grid(row=i + 1, column=0, sticky=tk.NSEW)

            self.inputs[galvo_labels[i] + " Amp"] = LabelInput(
                parent=self.galvo_frame,
                input_class=ValidatedSpinbox,
                input_var=tk.StringVar(),
            )

            self.inputs[galvo_labels[i] + " Amp"].grid(
                row=i + 1, column=1, sticky=tk.NSEW, pady=(10, 0), padx=(0, 5)
            )

            self.inputs[galvo_labels[i] + " Off"] = LabelInput(
                parent=self.galvo_frame,
                input_class=ValidatedSpinbox,
                input_var=tk.StringVar(),
            )

            self.inputs[galvo_labels[i] + " Off"].grid(
                row=i + 1, column=2, sticky=tk.NSEW, pady=(10, 0), padx=(0, 5)
            )

            self.inputs[galvo_labels[i] + " Rising"] = LabelInput(
                parent=self.galvo_frame,
                input_class=ValidatedSpinbox,
                input_var=tk.StringVar(),
                input_args={"from_": 0, "to": 100, "increment": 1}
            )

            self.inputs[galvo_labels[i] + " Rising"].grid(
                row=i + 1, column=3, sticky=tk.NSEW, pady=(10, 0), padx=(0, 5)
            )

            self.inputs[galvo_labels[i] + " Freq"] = LabelInput(
                parent=self.galvo_frame,
                input_class=ValidatedSpinbox,
                input_var=tk.StringVar(),
                input_args={"from_": 0, "to": 1000, "increment": 0.1},
            )

            self.inputs[galvo_labels[i] + " Freq"].grid(
                row=i + 1, column=4, sticky=tk.NSEW, pady=(10, 0), padx=(0, 5)
            )

            # Button for automatic estimate of galvo frequency
            self.buttons[galvo_labels[i] + " Freq"] = ttk.Button(
                self.galvo_frame, text="Estimate Frequency", padding=(2, 0, 2, 0)
            )

            self.buttons[galvo_labels[i] + " Freq"].grid(
                row=i + 1, column=5, sticky=tk.NE, pady=(10, 0)
            )

        self.inputs["galvo_info"] = LabelInput(
            parent=self.galvo_frame,
            input_class=ttk.Label,
            input_var=None,
            input_args={"text": "", "state": "disabled"},
        )
        self.inputs["galvo_info"].grid(
            row=len(galvo_labels) + 1, column=1, sticky=tk.NW, pady=(10, 0)
        )

        self.inputs["all_channels"] = LabelInput(
            parent=self.galvo_frame,
            input_class=ttk.Checkbutton,
            input_var=tk.BooleanVar(),
            label="Apply To All Channels",
        )
        self.inputs["all_channels"].grid(
            row=len(galvo_labels) + 1,
            column=3,
            columnspan=2,
            sticky=tk.NE,
            pady=(10, 0),
            padx=(0, 20),
        )
        self.inputs["all_channels"].widget.config(state="disabled")

        self.buttons["advanced_galvo_setting"] = ttk.Button(
            self.galvo_frame, text="Advanced Settings", padding=(5, 0, 5, 0)
        )
        self.buttons["advanced_galvo_setting"].grid(
            row=len(galvo_labels) + 1,
            column=5,
            sticky=tk.NSEW,
            pady=(10, 0),
        )

        # other remote focus waveform setting
        label = ttk.Label(self.waveform_frame, text="Remote Focus Settings")
        label.grid(row=0, columnspan=4, padx=5, sticky=tk.NW, pady=(10, 0))
        separator = ttk.Separator(self.waveform_frame)
        separator.grid(row=1, columnspan=4, sticky=tk.NSEW)
        rf_waveform_labels = [
            "Delay (ms)",
            "Fly Back Time (ms)",
            "Settle Duration (ms)",
            "Percent Smoothing",
        ]
        dict_labels = ["Delay", "Ramp_falling", "Duty", "Smoothing"]
        for i in range(len(dict_labels)):
            label = ttk.Label(
                self.waveform_frame, text=rf_waveform_labels[i], padding=(2, 5, 5, 0)
            )
            label.grid(
                row=i // 2 + 2,
                column=(i % 2) * 2,
                padx=(2 + (i % 2) * 40, 20),
                sticky=tk.NSEW,
            )
            self.inputs[dict_labels[i]] = LabelInput(
                parent=self.waveform_frame,
                input_class=ValidatedSpinbox,
                # label=rf_waveform_labels[i],
                input_var=tk.StringVar(),
                # label_args={"padding": (2, 5, 5, 0)},
                input_args={"from_": 0, "to": 100, "increment": 0.1},
            )
            self.inputs[dict_labels[i]].grid(
                row=i // 2 + 2, column=(i % 2) * 2 + 1, sticky=tk.NSEW
            )
        row_id = 2 + len(rf_waveform_labels) // 2
        label = ttk.Label(self.waveform_frame, text="Camera Settings")
        label.grid(row=row_id, column=0, padx=5, pady=(10, 0), sticky=tk.NW)
        separator = ttk.Separator(self.waveform_frame)
        separator.grid(row=row_id + 1, columnspan=4, sticky=tk.NSEW)
        camera_waveform_labels = ["Delay (ms)", "Settle Duration (ms)"]
        dict_labels = ["camera_delay", "camera_settle_duration"]
        for i in range(len(camera_waveform_labels)):
            label = ttk.Label(
                self.waveform_frame,
                text=camera_waveform_labels[i],
                padding=(2, 5, 5, 0),
            )
            label.grid(
                row=i // 2 + row_id + 2,
                column=(i % 2) * 2,
                padx=(2 + (i % 2) * 40, 20),
                sticky=tk.NSEW,
            )
            self.inputs[dict_labels[i]] = LabelInput(
                parent=self.waveform_frame,
                input_class=ValidatedSpinbox,
                # label=rf_waveform_labels[i],
                input_var=tk.StringVar(),
                # label_args={"padding": (2, 5, 5, 0)},
                input_args={"from_": 0, "to": 100, "increment": 0.1},
            )
            self.inputs[dict_labels[i]].grid(
                row=i // 2 + row_id + 2, column=(i % 2) * 2 + 1, sticky=tk.NSEW
            )

    # Getters
    def get_variables(self):
        """Returns the variables of the inputs

        This function returns a dictionary of all the variables
        that are tied to each widget name.
        The key is the widget name, value is the variable associated.

        Returns
        -------
        dict
            Dictionary of all the variables tied to each widget name
        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get_variable()
        return variables

    def get_widgets(self):
        """Returns the widgets

        This function returns the dictionary that holds the input widgets.
        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        dict
            Dictionary of all the widgets
        """
        return self.inputs

    def get_buttons(self):
        """Returns the buttons

        This function returns the dictionary that holds the buttons.
        The key is the button name, value is the button.

        Returns
        -------
        dict
            Dictionary of all the buttons
        """
        return self.buttons


class AdvancedWaveformParameterPopupWindow:
    """Popup window with advanced waveform parameters for galvos, remote focusing,
    etc."""

    def __init__(self, root, *args, **kwargs):
        """Initialize the AdvancedWaveformParameterPopupWindow

        Parameters
        ----------
        root : tk.Tk
            The root window
        """

        # Creating popup window with this name and size/placement, PopUp is a
        # Toplevel window

        #: PopUp: The popup window
        self.popup = PopUp(
            root, "Advanced Galvo Setting", "+320+180", top=False, transient=False
        )

        content_frame = self.popup.get_frame()
        # Formatting
        tk.Grid.columnconfigure(content_frame, "all", weight=1)
        tk.Grid.rowconfigure(content_frame, "all", weight=1)

        #: dict: Dictionary for all of the inputs
        self.inputs = {}

        #: dict: Dictionary for all the buttons
        self.buttons = {}

        #: dict: Dictionary for all the variables
        self.variables = {}

        #: dict: Dictionary for all the parameters
        self.parameters = {}

        #: ttk.Frame: Frame for top
        self.top_frame = ttk.Frame(content_frame, padding=(0, 0, 0, 0))

        #: ttk.Frame: Frame for parameters
        self.parameter_frame = ttk.Frame(content_frame, padding=(0, 0, 0, 0))

        self.top_frame.grid(row=0, column=0, sticky=tk.NSEW, padx=10, pady=(10, 0))
        self.parameter_frame.grid(
            row=1, column=0, sticky=tk.NSEW, padx=10, pady=(10, 0)
        )

        label = ttk.Label(self.top_frame, text="Galvo Parameters Associate with:")
        label.grid(row=0, column=0, columnspan=4, sticky=tk.NW)
        self.variables["galvo_factor"] = tk.StringVar()

        laser = ttk.Radiobutton(
            master=self.top_frame,
            text="Laser Wavelength",
            value="laser",
            variable=self.variables["galvo_factor"],
        )
        channel = ttk.Radiobutton(
            master=self.top_frame,
            text="Channel",
            value="channel",
            variable=self.variables["galvo_factor"],
        )
        none_factor = ttk.Radiobutton(
            master=self.top_frame,
            text="None",
            value="none",
            variable=self.variables["galvo_factor"],
        )
        self.top_frame.grid_columnconfigure(0, minsize=50)
        laser.grid(row=1, column=1, sticky=tk.NSEW, padx=5, pady=(10, 0))
        channel.grid(row=1, column=2, sticky=tk.NSEW, padx=5, pady=(10, 0))
        none_factor.grid(row=1, column=3, sticky=tk.NSEW, padx=5, pady=(10, 0))

    def generate_parameter_frame(self, factors=["All"], galvos=[]):
        """Generate galvo widgets

        Parameters
        ----------
        factors : list
            A list of galvo factor names
        galvos : list
            A list of galvo amplitude and offset values
        """
        # remove all widgets
        for child in self.parameter_frame.winfo_children():
            child.destroy()

        self.parameters = {}

        if len(galvos) < 3:
            self.parameter_frame.grid_columnconfigure(0, minsize=50)
            for i in range(len(galvos)):
                label = ttk.Label(self.parameter_frame, text=f"Galvo {i}")
                label.grid(row=0, column=i * 2 + 1, columnspan=2, padx=20)

                label = ttk.Label(self.parameter_frame, text="Amplitude")
                label.grid(
                    row=1, column=i * 2 + 1, sticky=tk.NSEW, padx=10, pady=(5, 0)
                )
                label = ttk.Label(self.parameter_frame, text="Offset")
                label.grid(
                    row=1, column=i * 2 + 2, sticky=tk.NSEW, padx=10, pady=(5, 0)
                )

            for i in range(len(factors)):
                label = ttk.Label(self.parameter_frame, text=factors[i])
                label.grid(row=i + 2, column=0, sticky=tk.NSEW, padx=20, pady=5)

                for j in range(len(galvos)):
                    self.parameters[f"galvo_{i}_{j}_amp"] = LabelInput(
                        parent=self.parameter_frame,
                        input_class=ValidatedSpinbox,
                        input_var=tk.StringVar(),
                    )
                    # self.parameters[f"galvo_{i}_{j}_amp"].set(galvos[j][i][0])
                    self.parameters[f"galvo_{i}_{j}_amp"].grid(
                        row=i + 2,
                        column=j * 2 + 1,
                        sticky=tk.NSEW,
                        padx=(0, 5),
                        pady=(0, 5),
                    )

                    self.parameters[f"galvo_{i}_{j}_off"] = LabelInput(
                        parent=self.parameter_frame,
                        input_class=ValidatedSpinbox,
                        input_var=tk.StringVar(),
                    )
                    # self.parameters[f"galvo_{i}_{j}_off"].set(galvos[j][i][1])
                    self.parameters[f"galvo_{i}_{j}_off"].grid(
                        row=i + 2,
                        column=j * 2 + 2,
                        sticky=tk.NSEW,
                        padx=(0, 5),
                        pady=(0, 5),
                    )
        else:
            # using tabs
            notebook = ttk.Notebook(self.parameter_frame)
            notebook.grid(row=0, column=0, padx=10, pady=10)

            for i in range(len(factors)):
                tab = tk.Frame(notebook, width=200)
                notebook.add(tab, text=factors[i])
                tab.grid_columnconfigure(0, minsize=50)
                label = ttk.Label(tab, text="Amplitude")
                label.grid(row=0, column=1, sticky=tk.NSEW, padx=10, pady=5)
                label = ttk.Label(tab, text="Offset")
                label.grid(row=0, column=2, sticky=tk.NSEW, padx=10, pady=5)

                for j in range(len(galvos)):
                    label = ttk.Label(tab, text=f"Galvo {j}")
                    label.grid(
                        row=j + 1, column=0, sticky=tk.NSEW, padx=10, pady=(0, 5)
                    )

                    self.parameters[f"galvo_{i}_{j}_amp"] = LabelInput(
                        parent=tab,
                        input_class=ValidatedSpinbox,
                        input_var=tk.StringVar(),
                    )
                    # self.parameters[f"galvo_{i}_{j}_amp"].set(galvos[j][i][0])
                    self.parameters[f"galvo_{i}_{j}_amp"].grid(
                        row=j + 1, column=1, sticky=tk.NSEW, padx=10, pady=(0, 5)
                    )
                    self.parameters[f"galvo_{i}_{j}_off"] = LabelInput(
                        parent=tab,
                        input_class=ValidatedSpinbox,
                        input_var=tk.StringVar(),
                    )
                    # self.parameters[f"galvo_{i}_{j}_off"].set(galvos[j][i][1])
                    self.parameters[f"galvo_{i}_{j}_off"].grid(
                        row=j + 1, column=2, sticky=tk.NSEW, padx=10, pady=(0, 5)
                    )



"""
Stage devices.
"""

device_types = {
    "asi": "Applied Scientific Instrumentation",
    "galvo": "Analog/Digital Device",
    "mcl": "Mad City Labs",
    "pi": "Physik Instrumente",
    "sutter": "Sutter Instruments",
    "synthetic": "Virtual Device",
    "tl_kcube_inertial": "ThorLabs KCube Inertial Device",
}



"""
Camera devices.
"""

device_types = {
    "hamamatsu_lightning": "Hamamatsu ORCA Lightning",
    "hamamatsu_fire": "Hamamatsu ORCA Fire",
    "hamamatsu_flash": "Hamamatsu Flash 4.0",
    "photometics": "Photometrics Iris 15B",
    "synthetic": "Virtual Device",
}



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

#  Standard Library Imports
import logging
import time

# Third Party Imports

# Local Imports
from navigate.model.devices.filter_wheel.base import FilterWheelBase
from navigate.model.devices.device_types import SerialDevice
from navigate.model.devices.APIs.asi.asi_tiger_controller import TigerController
from navigate.tools.decorators import log_initialization

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


@log_initialization
class ASIFilterWheel(FilterWheelBase, SerialDevice):
    """ASIFilterWheel - Class for controlling ASI Filter Wheels

    Note
    ----
        Additional information on the ASI Filter Wheel can be found at:
        https://asiimaging.com/docs/fw_1000#fw-1000_ascii_command_set
    """

    def __init__(self, microscope_name, device_connection, configuration, device_id=0):
        """Initialize the ASIFilterWheel class.

        Parameters
        ----------
        microscope_name : str
            Name of the microscope.
        device_connection : TigerController
            Communication object for the ASI Filter Wheel.
        configuration : dict
            Global configuration of the microscope.
        device_id : int
            The ID of the device. Default is 0.
        """

        super().__init__(microscope_name, device_connection, configuration, device_id)

        #: float: Delay for filter wheel to change positions.
        self.wait_until_done_delay = self.device_config["filter_wheel_delay"]

        self.filter_wheel = device_connection

        # Send Filter Wheel/Wheels to Zeroth Position
        self.filter_wheel.select_filter_wheel(
            filter_wheel_number=self.filter_wheel_number
        )

        self.filter_wheel.move_filter_wheel(filter_wheel_position=0)

        #: int: Filter wheel position.
        self.filter_wheel_position = 0

    def __str__(self):
        """String representation of the class."""
        return "ASIFilterWheel"

    @classmethod
    def connect(cls, port, baudrate=115200, timeout=0.25):
        """Build ASIFilterWheel Serial Port connection

        Parameters
        ----------
        port : str
            Port for communicating with the filter wheel, e.g., COM1.
        baudrate : int
            Baud rate for communicating with the filter wheel, default is 115200.
        timeout : float
            Timeout for communicating with the filter wheel, default is 0.25.

        Returns
        -------
        tiger_controller : TigerController
            ASI Tiger Controller object.
        """
        # wait until ASI device is ready
        tiger_controller = TigerController(port, baudrate)
        tiger_controller.connect_to_serial()
        if not tiger_controller.is_open():
            logger.error("ASI stage connection failed.")
            raise Exception("ASI stage connection failed.")
        return tiger_controller

    def filter_change_delay(self, filter_name):
        """Estimate duration of time necessary to move the filter wheel

        Assumes that it is ~40ms per adjacent position.
        Depends on filter wheel parameters and load.

        Parameters
        ----------
        filter_name : str
            Name of filter to move to.

        """
        old_position = self.filter_wheel_position
        new_position = self.filter_dictionary[filter_name]
        delta_position = int(abs(old_position - new_position))
        self.wait_until_done_delay = delta_position * 0.04

    def set_filter(self, filter_name, wait_until_done=True):
        """Change the filter wheel to the filter designated by the filter
        position argument.

        Parameters
        ----------
        filter_name : str
            Name of filter to move to.
        wait_until_done : bool
            Waits duration of time necessary for filter wheel to change positions.
        """
        if self.check_if_filter_in_filter_dictionary(filter_name) is True:

            # Calculate the Delay Needed to Change the Positions
            self.filter_change_delay(filter_name)

            self.filter_wheel.select_filter_wheel(
                filter_wheel_number=self.filter_wheel_number
            )
            self.filter_wheel.move_filter_wheel(self.filter_dictionary[filter_name])

            #  Wheel Position Change Delay
            if wait_until_done:
                time.sleep(self.wait_until_done_delay)

    def close(self):
        """Close the ASI Filter Wheel serial port.

        Sets the filter wheel to the home position and then closes the port.
        """
        if self.filter_wheel.is_open():
            self.filter_wheel.move_filter_wheel_to_home()
            logger.debug("ASI Filter Wheel - Closing Device.")
            self.filter_wheel.disconnect_from_serial()

    def __del__(self):
        """Destructor for the ASIFilterWheel class."""
        self.close()


@log_initialization
class ASICubeSliderFilterWheel(FilterWheelBase, SerialDevice):
    """ASICubeSlider - Class for controlling the C60 Cube Slider from ASI.

    Note
    ----
        Additional information on the ASI Filter Wheel can be found at:
        www.asiimaging.com/docs/filter_and_turret_changer?s%5B%5D=filter&s%5B%5D=slider

        Maximum number of positions is 4.

        Typical switch time between adjacent positions is < 250 ms.
    """

    def __init__(self, microscope_name, device_connection, configuration, device_id=0):
        """Initialize the ASICubeSlider class.

        Parameters
        ----------
        microscope_name : str
            Name of the microscope.
        device_connection : TigerController
            Communication object for the ASI Filter Wheel.
        configuration : dict
            Global configuration of the microscope.
        device_id : int
            The ID of the device. Default is 0.
        """

        super().__init__(microscope_name, device_connection, configuration)

        #: obj: ASI Tiger Controller object.
        self.dichroic = device_connection

        #: float: Delay for filter wheel to change positions.
        self.wait_until_done_delay = self.device_config["filter_wheel_delay"]

        #: str: The ID of the dichroic in the Tiger Controller. e.g., "T"
        self.dichroic_id = self.filter_wheel_number

        self.dichroic.move_dichroic(dichroic_id=self.dichroic_id, dichroic_position=0)

        #: int: Filter wheel position.
        self.dichroic_position = 0

    @classmethod
    def connect(cls, port, baudrate=115200, timeout=0.25):
        """Build ASIFilterWheel Serial Port connection

        Parameters
        ----------
        port : str
            Port for communicating with the filter wheel, e.g., COM1.
        baudrate : int
            Baud rate for communicating with the filter wheel, default is 115200.
        timeout : float
            Timeout for communicating with the filter wheel, default is 0.25.

        Returns
        -------
        tiger_controller : TigerController
            ASI Tiger Controller object.
        """
        # wait until ASI device is ready
        tiger_controller = TigerController(port, baudrate)
        tiger_controller.connect_to_serial()
        if not tiger_controller.is_open():
            logger.error("ASI stage connection failed.")
            raise Exception("ASI stage connection failed.")
        return tiger_controller

    def filter_change_delay(self, filter_name):
        """Estimate duration of time necessary to move the dichroic

        Assumes that it is <250 ms per adjacent position.

        Parameters
        ----------
        filter_name : str
            Name of filter to move to.

        """
        old_position = self.dichroic_position
        new_position = self.filter_dictionary[filter_name]
        delta_position = int(abs(old_position - new_position))
        self.wait_until_done_delay = delta_position * 0.25

    def set_filter(self, filter_name, wait_until_done=True):
        """Change the dichroic position.

        Parameters
        ----------
        filter_name : str
            Name of filter to move to.
        wait_until_done : bool
            Waits duration of time necessary for filter wheel to change positions.
        """
        if self.check_if_filter_in_filter_dictionary(filter_name) is True:

            # Calculate the Delay Needed to Change the Positions
            self.filter_change_delay(filter_name)
            dichroic_position = self.filter_dictionary[filter_name]

            assert dichroic_position in range(4)
            self.dichroic.move_dichroic(
                dichroic_id=self.dichroic_id, dichroic_position=dichroic_position
            )

            #  Wheel Position Change Delay
            if wait_until_done:
                time.sleep(self.wait_until_done_delay)

    def close(self):
        """Close the ASI Filter Wheel serial port.

        Sets the filter wheel to the home position and then closes the port.
        """
        if self.dichroic.is_open():
            logger.debug("ASI C60 Motorized Slider - Closing Device.")
            self.dichroic.disconnect_from_serial()



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard library imports
from typing import Optional, Union, List, Dict
import logging

# Third-party imports
import numpy.typing as npt

# Local application imports
from .metadata import Metadata

NGFF_VERSION = "0.4"

p = __name__.split(".")[1]
logger = logging.getLogger(p)


class OMEZarrMetadata(Metadata):
    """Class to generate OME-Zarr metadata."""

    @property
    def _axes(self) -> Dict:
        """Return tczyx axes in navigate units.

        https://ngff.openmicroscopy.org/0.4/#axes-md

        Returns
        -------
        List
            A list of dictionaries containing the axis name, type, and unit.
        """
        axes = [
            {"name": "t", "type": "time", "unit": "second"},
            {"name": "c", "type": "channel"},
            {"name": "z", "type": "space", "unit": "micrometer"},
            {"name": "y", "type": "space", "unit": "micrometer"},
            {"name": "x", "type": "space", "unit": "micrometer"},
        ]

        return axes

    def _stage_positions_to_translation_transform(
        self, x: float, y: float, z: float, theta: float, f: Optional[float] = None
    ) -> List[float]:
        """Grab the translation from the stage.

        Ignore theta, focus for now.

        Parameters
        ----------
        x : float
            The x position of the stage (um).
        y : float
            The y position of the stage (um).
        z : float
            The z position of the stage (um).
        theta : float
            The theta position of the stage (deg).
        f : float
            The focus position of the stage (um).

        Returns
        -------
        List
            A translation for each axis.
        """

        # Set the transform positions
        xp, yp, zp = x, y, z

        # Allow additional axes (e.g. f) to couple onto existing axes (e.g. z)
        # if they are both moving along the same physical dimension
        if self._coupled_axes is not None:
            for leader, follower in self._coupled_axes.items():
                if leader.lower() not in "xyz":
                    print(
                        f"Unrecognized coupled axis {leader}. "
                        "Not gonna do anything with this."
                    )
                    continue
                elif leader.lower() == "x":
                    xp += float(locals()[f"{follower.lower()}"])
                elif leader.lower() == "y":
                    yp += float(locals()[f"{follower.lower()}"])
                elif leader.lower() == "z":
                    zp += float(locals()[f"{follower.lower()}"])

        # t, c, z, y, x
        return [0.0, 0.0, zp, yp, xp]

    def _scale_transform(
        self, subdiv: Union[npt.ArrayLike, List] = [1, 1, 1]
    ) -> List[float]:
        """Calculate the image scale after subdivision.

        Parameters
        ----------
        subdiv : List
            The subdivision of the dataset.

        Returns
        -------
        List
            The scale of the dataset.
        """
        return [
            1,  # t
            1,  # c
            self.dz * subdiv[2],
            self.dy * subdiv[1],
            self.dx * subdiv[0],
        ]

    def _coordinate_transformations(
        self, scale: Optional[List] = None, translation: Optional[List] = None
    ) -> Dict:
        """Package scale and translation in a dict.

        Parameters
        ----------
        scale : List
            The scale of the dataset.
        translation : List
            The translation of the dataset.

        Returns
        -------
        Dict
            A dictionary containing the transformation.
        """
        transformation = []
        if (
            scale is None
            or not isinstance(scale, list)
            or len(scale) != len(self._axes)
        ):
            if translation is not None:
                logger.error("Translation cannot be provided without scale.")
                raise UserWarning("Translation cannot be provided without scale.")
            return transformation
        else:
            transformation.append({"type": "scale", "scale": scale})

        if (
            translation is not None
            and isinstance(translation, list)
            and len(translation) == len(self._axes)
        ):
            transformation.append({"type": "translation", "translation": translation})

        return transformation

    def multiscales_dict(
        self,
        name: str,
        paths: list,
        resolutions: Union[npt.ArrayLike, List],
        view: Optional[Dict] = None,
    ) -> Dict:
        """Create a multiscale dictionary for the OME-Zarr metadata.

        https://ngff.openmicroscopy.org/0.4/index.html#multiscale-md

        Parameters
        ----------
        name : str
            The name of the dataset.
        paths : list
            The paths of the dataset.
        resolutions : List
            The resolutions of the dataset.
        view : Dict
            The view of the dataset.

        Returns
        -------
        Dict
            A dictionary containing the multiscale metadata.
        """
        d = {"version": NGFF_VERSION, "name": name, "axes": self._axes}

        datasets = []
        for path, res in zip(paths, resolutions):
            scale = self._scale_transform(res)
            dd = {
                "path": path,
                "coordinateTransformations": self._coordinate_transformations(scale),
            }
            datasets.append(dd)
        d["datasets"] = datasets

        if view is not None:
            scale = [1.0] * len(self._axes)
            translation = self._stage_positions_to_translation_transform(**view)
            d["coordinateTransformations"] = self._coordinate_transformations(
                scale, translation
            )

        return d



# Copyright (c) 2021-2025  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

from abc import ABC, abstractmethod

class DeviceBase(ABC):
    """DeviceBase - Parent device class."""

    def __init__(
        self,
        device_name: str,
        *args,
        **kwargs,
    ):
        """Initialize DeviceBase class.
        """

        self.device_name = device_name

        self.unique_id = device_name

        self.device_connection = None

    @abstractmethod
    def connect(self):
        pass


class SerialDevice:
    """SerialDevice - Parent serial device class."""
    
    def __init__(
        self,
        device_name: str,
        port: str = "",
        baudrate=115200,
        timeout=0.25,
        **kwargs,
    ):
        """Initialize SerialDevice class.
        """

        self.device_name = device_name
        self.unique_id = "serial_" + port

    def connect(self, port, baudrate=115200, timeout=0.25):
        """Connect to serial device.
        """
        if port:
            from serial import Serial
            self.serial = Serial()
            self.serial.port = port
            self.serial.baudrate = baudrate
            self.serial.timeout = timeout
            self.serial.open()
        else:
            self.serial = None

        return self.serial

    def disconnect(self):
        """Disconnect from serial device.
        """
        try:
            if self.serial.is_open:
                self.serial.close()
        except Exception as e:
            print(f"Error disconnecting from serial device: {e}")

class IntegratedDevice:
    """IntegratedDevice - Parent integrated device class."""

class NIDevice:
    """NIDevice - Parent National Instruments device class."""

class SequenceDevice:
    """SequenceDevice - The device loaded according to its sequence id, not serial number.
    Always need to check if the serial number is match.
    """



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


def update_nested_dict(d, find_func, apply_func):
    """Update a nested dictionary by applying a function to a value

    Loops through a nested dictionary and if find_func() conditions are met,
    run apply_func on that key.

    TODO: It might be nice to make this non-recursive.

    Parameters
    ----------
    d : dict
        Dictionary to be updated
    find_func : func
        Accepts key, value pair and matches a condition based on these. Returns bool.
    apply_func : func
        Accepts a value returns the new value.

    Returns
    -------
    d2 : dict
        An version of d, updated according to the passed functions.
    """
    d2 = {}
    for k, v in d.items():
        if find_func(k, v):
            d2[k] = apply_func(v)
        else:
            d2[k] = v
        if isinstance(v, dict):
            d2[k] = update_nested_dict(v, find_func, apply_func)
    return d2


def update_stage_dict(target, pos_dict):
    """Update dictionary entries common to the model and controller.

    Parameters
    ----------
    target : navigate.model.Model or navigate.controller.Controller
        The object that is being updated.
    pos_dict : dict
        The dictionary of positions to update.

    Returns
    -------
    None
    """
    # Update our local experiment parameters
    for axis, val in pos_dict.items():
        ax = axis.split("_")[0]
        target.configuration["experiment"]["StageParameters"][ax] = val



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
from typing import Dict, Any
import tkinter as tk
from tkinter import ttk

# Third Party Imports

# Local Imports


def uniform_grid(cls: Any) -> None:
    """This function is used to equally distribute the columns and rows of a
    tkinter frame.

    Parameters
    ----------
    cls : tk.Frame or ttk.Frame
        The class that is to be distributed.
    """
    cols, rows = cls.grid_size()
    for col in range(cols):
        cls.grid_columnconfigure(col, weight=1)
    for row in range(rows):
        cls.grid_rowconfigure(row, weight=1)

class CommonMethods:
    """This class is a collection of common methods for handling variables, widgets,
    and buttons.
    """

    def get_variables(self) -> Dict[str, Any]:
        """This function returns a dictionary of all the variables that are tied to
        each  widget name.

        The key is the widget name, value is the variable associated.

        Returns
        -------
        variables : dict
            The dictionary that holds the variables.
        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get()
        return variables

    def get_widgets(self) -> Dict[str, Any]:
        """This function returns the dictionary that holds the widgets.

        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        widgets : dict
            The dictionary that holds the widgets.
        """
        return self.inputs

    def get_buttons(self) -> Dict[str, ttk.Button]:
        """Get the buttons of the popup

        This function returns the dictionary that holds the buttons.
        The key is the button name, value is the button.

        Returns
        -------
        buttons : Dict[str, ttk.Button]
            Dictionary of all the buttons
        """
        return self.buttons



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

#  Standard Library Imports
import logging
import time

# Third Party Imports
import numpy as np
import serial

# Local Imports
from navigate.model.devices.filter_wheel.base import FilterWheelBase
from navigate.model.devices.device_types import SerialDevice
from navigate.tools.decorators import log_initialization

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


@log_initialization
class SutterFilterWheel(FilterWheelBase, SerialDevice):
    """SutterFilterWheel - Class for controlling Sutter Lambda Filter Wheels

    Note
    ----
        Additional information on the Sutter Lambda 10-B can be found at:
        https://www.sutter.com/manuals/LB10-3_OpMan.pdf
    """

    def __init__(self, microscope_name, device_connection, configuration, device_id):
        """Initialize the SutterFilterWheel class.

        Parameters
        ----------
        microscope_name : str
            Name of the microscope.
        device_connection : Any
            The communication instance with the device.
        configuration : Dict[str, Any]
            Global configuration dictionary.
        device_id : int
            The ID of the device. Default is 0.
        """
        super().__init__(microscope_name, device_connection, configuration, device_id)

        #: bool: Wait until filter wheel has completed movement.
        self.wait_until_done = True

        #: float: Delay in s for the wait until done function.
        self.wait_until_done_delay = None

        #: bool: Read on initialization.
        self.read_on_init = True

        #: int: Filter wheel speed.
        self.speed = 2

        # Delay in s for the wait until done function
        #: np.matrix: Delay matrix for filter wheel.
        self.delay_matrix = np.array(
            [
                [0, 0.031, 0.051, 0.074, 0.095, 0.115],
                [0, 0.040, 0.065, 0.095, 0.120, 0.148],
                [0, 0.044, 0.075, 0.105, 0.136, 0.168],
                [0, 0.050, 0.088, 0.127, 0.165, 0.205],
                [0, 0.060, 0.108, 0.156, 0.205, 0.250],
                [0, 0.068, 0.123, 0.178, 0.235, 0.290],
                [0, 0.124, 0.235, 0.350, 0.460, 0.580],
                [0, 0.230, 0.440, 0.650, 0.860, 1.100],
            ]
        )
        self.serial = device_connection
        self.serial.write(bytes.fromhex("ee"))

        if self.read_on_init:
            self.read(2)  # class 'bytes'
            #: bool: Software initialization complete flag.
            self.init_finished = True
        else:
            self.init_finished = False

        # Set filter to the 0th position by default upon initialization.
        self.set_filter(list(self.filter_dictionary.keys())[0])

    def __str__(self) -> str:
        """String representation of the class."""
        return "SutterFilterWheel"

    @classmethod
    def connect(cls, comport, baudrate, timeout=0.25):
        """Build SutterFilterWheel Serial Port connection

        Attributes
        ----------
        comport : str
            Comport for communicating with the filter wheel, e.g., COM1.
        baudrate : int
            Baud rate for communicating with the filter wheel, e.g., 9600.
        timeout : float
            Timeout for communicating with the filter wheel, e.g., 0.25.

        Returns
        -------
        serial.Serial
            The serial port connection to the filter wheel.

        Raises
        ------
        UserWarning
            Could not communicate with Sutter Lambda 10-B via COMPORT.
        """
        logging.debug(f"SutterFilterWheel - Opening Serial Port {comport}")
        try:
            return serial.Serial(comport, baudrate, timeout=timeout)
        except serial.SerialException:
            logger.error("SutterFilterWheel - Could not establish Serial Port Connection")
            raise UserWarning(
                "Could not communicate with Sutter Lambda 10-B via COMPORT", comport
            )

    def filter_change_delay(self, filter_name: str) -> None:
        """Calculate duration of time necessary to change filter wheel positions.

        Calculate duration of time necessary to change filter wheel positions.
        Identifies the number of positions that must be switched, and then retrieves
        the duration of time necessary to perform the switch from self.delay_matrix.

        Parameters
        ----------
        filter_name : str
            Name of filter that we want to move to.

        Note
        ----
            Detailed information on timing located on page 38:
            https://www.sutter.com/manuals/LB10-3_OpMan.pdf

        Warn
        ----
            Delay matrix should be model specific.

        Returns
        -------
        Index Error
            If the filter wheel position is out of range.
        """
        # Find the old and new positions, the distance between them, and the
        # delay necessary.
        old_position = self.wheel_position
        self.wheel_position = self.filter_dictionary[filter_name]
        delta_position = int(abs(old_position - self.wheel_position))
        try:
            self.wait_until_done_delay = self.delay_matrix[self.speed, delta_position]
        except IndexError:
            self.wait_until_done_delay = 0.01

    def set_filter(self, filter_name: str, wait_until_done: bool = True):
        """Change the filter wheel to the filter designated by the filter
        position argument.

        Parameters
        ----------
        filter_name : str
            Name of filter to move to.
        wait_until_done : bool
            Waits duration of time necessary for filter wheel to change positions.
        """
        if self.check_if_filter_in_filter_dictionary(filter_name) is True:

            # Calculate the Delay Needed to Change the Positions
            self.filter_change_delay(filter_name)

            # Make sure you are moving it to a reasonable filter position at a
            # reasonable speed.
            assert self.wheel_position in range(10)
            assert self.speed in range(8)

            # If previously we did not confirm that the initialization was complete,
            # check now.
            if not self.init_finished:
                self.read(2)
                self.init_finished = True
                logger.debug("SutterFilterWheel - Initialized.")

            """send the binary sequence via serial to move to the desired
            filter wheel position
            When number_of_filter_wheels = 1, wheel A changes.
            When number_of_filter_wheels = 2, wheel B changes.
            Filter Wheel Command Byte Encoding = wheel + (self.speed*16) +
            position = command byte
            """
            logger.debug(
                f"SutterFilterWheel - Moving to Position {self.filter_wheel_number-1}"
            )
            output_command = (
                (self.filter_wheel_number - 1) * 128
                + self.wheel_position
                + 16 * self.speed
            )
            output_command = output_command.to_bytes(1, "little")
            self.serial.write(output_command)
            # read echoing back
            self.read(1)

            #  Wheel Position Change Delay
            if wait_until_done:
                time.sleep(self.wait_until_done_delay)
                # read 0D back.
                self.read(1)

    def read(self, num_bytes: int) -> bytes:
        """Reads the specified number of bytes from the serial port.

        Parameters
        ----------
        num_bytes : int
            Number of bytes to read from the serial port.

        Returns
        -------
        response : bytes
            The bytes read from the serial port.

        Raises
        ------
        UserWarning
            The serial port to the Sutter Lambda 10-B is on, but it isn't responding as
            expected.
        """
        for i in range(100):
            num_waiting = self.serial.inWaiting()
            # if there are unread returns from previous commands
            if num_waiting >= num_bytes:
                break
            time.sleep(0.02)
        else:
            logger.error(
                "The serial port to the Sutter Lambda 10-B is on, but it isn't "
                "responding as expected."
            )
            raise UserWarning(
                "The serial port to the Sutter Lambda 10-B is on, but it isn't "
                "responding as expected."
            )
        return self.serial.read(num_bytes)

    def close(self) -> None:
        """Close the SutterFilterWheel serial port.

        Sets the filter wheel to the Empty-Alignment position and then closes the port.
        """
        logger.debug("SutterFilterWheel - Closing the Filter Wheel Serial Port")
        self.set_filter(list(self.filter_dictionary.keys())[0])
        self.serial.close()

    def __del__(self) -> None:
        """Delete the SutterFilterWheel class."""
        if self.serial.is_open:
            self.close()



# Copyright (c) 2021-2025  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import logging
from typing import Any, Dict, Optional

# Third Party Imports
from ximea import xiapi

# Local Imports
from navigate.model.devices.camera.base import CameraBase
from navigate.tools.decorators import log_initialization

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


@log_initialization
class XimeaBase(CameraBase):
    """Ximea camera base class.
    """

    def __init__(
        self,
        microscope_name: str,
        device_connection: Any,
        configuration: Dict[str, Any],
        *args: Optional[Any],
        **kwargs: Optional[Any],
    ) -> None:
        """Initialize MU196XR class.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_connection : Any
            Hardware device to connect to
        configuration : Dict[str, Any]
            Global configuration of the microscope
        """
        super().__init__(microscope_name, device_connection, configuration)

        #: str: Name of the microscope
        self.microscope_name = microscope_name

        #: object: Camera Object
        self.cam = device_connection

        #: dict: Configuration settings
        self.configuration = configuration

        #: dict: Camera parameters
        self.camera_parameters["x_pixels"] = self.cam.get_param("width:max")
        self.camera_parameters["y_pixels"] = self.cam.get_param("height:max")
        self.camera_parameters["x_pixels_min"] = self.cam.get_param("width:min")
        self.camera_parameters["y_pixels_min"] = self.cam.get_param("height:min")
        self.camera_parameters["x_pixels_step"] = self.cam.get_param("width:inc")
        self.camera_parameters["y_pixels_step"] = self.cam.get_param("height:inc")
        self.offset_x_min = self.cam.get_param("offsetX:min")
        self.offset_x_step = self.cam.get_param("offsetX:inc")
        self.offset_y_min = self.cam.get_param("offsetY:min")
        self.offset_y_step = self.cam.get_param("offsetY:inc")
        self.minimum_exposure_time = self.cam.get_param("exposure:min")

        # set external rising edge trigger: XI_TRG_EDGE_RISING
        input_trigger_port = int(self.camera_parameters.get("input_trigger_port", 1))
        if input_trigger_port >= 1 and input_trigger_port <= 12:
            input_trigger_port = f"XI_GPI_PORT{input_trigger_port}"
        else:
            input_trigger_port = "XI_GPI_PORT1"
        self.cam.set_param("gpi_selector", input_trigger_port)
        self.cam.set_param("gpi_mode", "XI_GPI_TRIGGER")
        self.cam.set_param('trigger_source', "XI_TRG_EDGE_RISING")

        self.camera_parameters["supported_sensor_modes"] = ["Normal"]
        self.camera_parameters["supported_readout_directions"] = ["Top-to-Bottom"]


    def __str__(str):
        """Return string representation of Ximea Base class
        
        Returns
        -------
        str
            String representation of Ximea Base class.
        """
        return "XimeaBase"
    
    def __del__(self):
        """Delete Ximea Camera class."""
        self.cam.close_device()

    @classmethod
    def get_connect_params(cls):
        """Register the parameters required to connect to the camera.

        Returns
        -------
        list
            List of parameters required to connect to the camera.
        """
        return ["serial_number"]

    @classmethod
    def connect(cls, serial_number):
        """Build Photometrics Stage Serial Port connection

        Import Photometrics API and Initialize Camera Controller.

        Parameters
        ----------
        serial_number : str
            Camera serial number

        Returns
        -------
        cam : object
            Camera object.
        """
        try:
            cam = xiapi.Camera()
            # XI_OPEN_BY_SN
            cam.open_device_by_SN(str(serial_number))
            return cam
        except Exception as e:
            logger.error(f"Could not establish connection with camera: {e}.")
            raise UserWarning(
                "Could not establish connection with XIMEA camera", serial_number
            )
        
    @property
    def serial_number(self):
        """Get Camera Serial Number

        Returns
        -------
        serial_number : str
            Serial number for the camera.
        """
        self.cam.get_param("device_sn").decode("utf-8")

    def report_settings(self):
        """Print Camera Settings.

        Prints the current camera settings to the console and the log file.
        """
        params = ["exposure", "downsampling", "width", "height", "offsetX", "offsetY", "shutter_type"]
        for param in params:
            value = self.cam.get_param(param)
            print(param, value)
            logger.info(f"{param}, {value}")

    def set_sensor_mode(self, mode):
        """Set Ximea sensor mode.

        On the manual page 72:
        Cameras can be operated in two shutter modes, Rolling Shutter or Global Reset Release. 
        The Rolling Shutter mode is used if the camera is operated in free-run mode. 
        If the camera is triggered, either by hardware trigger or through software, 
        the sensor uses the Global Reset Release mode.

        Confirmed that the camera can run in free-run mode even if it is triggered by software.

        Parameters
        ----------
        mode : str
            'Normal' or 'Light-Sheet'
        """
        self.cam.set_param("acq_timing_mode", "XI_ACQ_TIMING_MODE_FREE_RUN")
        self.cam.set_param("shutter_type", "XI_SHUTTER_ROLLING")
    
    def set_readout_direction(self, mode):
        """Set readout direction"""
        pass

    def calculate_readout_time(self):
        """Get the duration of time needed to read out an image.

        Returns
        -------
        readout_time : float
            Duration of time needed to read out an image.
        """
        return 0
    
    def set_exposure_time(self, exposure_time):
        """Set Ximea exposure time.

        Note
        ----
            Units of the Ximea API are in microseconds.

        Parameters
        ----------
        exposure_time : float
            Exposure time in seconds.
        """
        #seconds to us.
        self.cam.set_param("exposure", exposure_time * 1000000)
        return True
    
    def set_line_interval(self, line_interval_time):
        """Set line interval.

        Parameters
        ----------
        line_interval_time : float
            Line interval duration.
        """
        return False

    def get_line_interval(self):
        """Get line interval.

        Returns
        -------
        line_interval_time : float
            Line interval duration.
        """
        self.line_interval = 0

    def set_binning(self, binning_string):
        """Set Ximea Camera binning mode.

        Parameters
        ----------
        binning_string : str
            Desired binning properties (e.g., '1x1', '2x2', '4x4', '8x8', '16x16',
            '1x2', '2x4')

        Returns
        -------
        result: bool
            True if successful, False otherwise.
        """
        binning_dict = {
            "1x1": 1,
            "2x2": 2,
            "3x3": 3,
            "4x4": 4,
            "5x5": 5,
            "6x6": 6,
            "7x7": 7,
            "8x8": 8,
            "9x9": 9,
            "10x10": 10,
            "16x16": 16
        }
        if binning_string not in binning_dict.keys():
            logger.debug(f"can't set binning to {binning_string}")
            print(f"can't set binning to {binning_string}")
            return False
        self.cam.set_param("downsampling_type", "XI_BINNING") # XI_BINNING
        self.cam.set_param("downsampling", "XI_DWN_" + binning_string)
        # after setting downsampling, the image size changes (width, height, imgpayloadsize)

        return True

    def set_ROI(self, roi_width=2048, roi_height=2048, center_x=1024, center_y=1024):
        """Change the size of the active region on the camera.

        Parameters
        ----------
        roi_width : int
            Width of active camera region.
        roi_height : int
            Height of active camera region.
        center_x : int
            X position of the center of view
        center_y : int
            Y position of the center of view

        Returns
        -------
        result: bool
            True if successful, False otherwise.
        """
        binning_value = int(self.cam.get_param("downsampling")[len("XI_DWN_"):].split("x")[0])

        roi_width = roi_width // binning_value
        roi_width = roi_width - roi_width % self.camera_parameters["x_pixels_step"]
        roi_height = roi_height // binning_value
        roi_height = roi_height - roi_height % self.camera_parameters["y_pixels_step"]
        offset_x = center_x - (roi_width * binning_value) // 2
        offset_x = offset_x - offset_x % self.offset_x_step
        offset_y = center_y - (roi_height * binning_value) // 2
        offset_y = offset_y - offset_y % self.offset_y_step

        x_max = self.camera_parameters["x_pixels"] // binning_value
        y_max = self.camera_parameters["y_pixels"] // binning_value

        if (roi_width % self.camera_parameters["x_pixels_step"] != 0
            or roi_height % self.camera_parameters["y_pixels_step"] != 0
            or offset_x < self.offset_x_min
            or offset_x % self.offset_x_step != 0
            or offset_y < self.offset_y_min
            or offset_y % self.offset_y_step != 0
            or offset_x + roi_width > x_max
            or offset_y + roi_height > y_max
        ):
            logger.debug(f"can't set roi to {roi_width} and {roi_height} with the center ({center_x}, {center_y})")
            return False
        
        # width and height are actual image size, not the ROI size in Ximea Camera
        try:
            self.cam.set_param("width", roi_width)
            self.cam.set_param("height", roi_height)
            self.cam.set_param("offsetX", offset_x)
            self.cam.set_param("offsetY", offset_y)
        except xiapi.Xi_error as e:
            logger.error(f"Error setting ROI: {e} with {roi_width} and {roi_height} with the offset ({offset_x}, {offset_y})")
            return False

        self.x_pixels = self.cam.get_param("width") * binning_value
        self.y_pixels = self.cam.get_param("height") * binning_value

        return self.x_pixels == roi_width * binning_value and self.y_pixels == roi_height * binning_value
    
    def set_ROI_and_binning(self, roi_width=2048, roi_height=2048, center_x=1024, center_y=1024, binning='1x1'):
        """Change the size of the active region on the camera and set the binning mode.

        Parameters
        ----------
        roi_width : int
            Width of active camera region.
        roi_height : int
            Height of active camera region.
        center_x : int
            X position of the center of view
        center_y : int
            Y position of the center of view
        binning : str
            Desired binning properties (e.g., '1x1', '2x2', '4x4', '8x8', '16x16',
            '1x2', '2x4')
        
        Returns
        -------
        result: bool
            True if successful, False otherwise.
        """
        # Set Binning
        result = self.set_binning(binning)
        if not result:
            return False
        
        # Set ROI
        result = self.set_ROI(roi_width, roi_height, center_x, center_y)
        return result

    def initialize_image_series(self, data_buffer=None, number_of_frames=100):
        """Initialize Ximea Camera image series.

        Parameters
        ----------
        data_buffer :
            List of SharedNDArrays of shape=(self.img_height,
            self.img_width) and dtype="uint16"
            Default is None.
        number_of_frames : int
            Number of frames.  Default is 100.
        """
        self._data_buffer = data_buffer
        self._number_of_frames = number_of_frames
        self._frames_received = 0

        # set buffer policy: XI_BP_SAFE
        self.cam.set_param("buffer_policy", "XI_BP_SAFE")
        # set image data format to XI_MONO16, this value can be set only if acquisition is stopped.
        self.cam.set_param('imgdataformat', "XI_MONO16")
        #imgpayloadsize changes automatically after setting imgdataformat
        # self.cam.get_param("imgpayloadsize")
        # start_acquisition
        self._image = xiapi.Image()
        self.cam.start_acquisition()
        self.is_acquiring = True

    def close_image_series(self):
        """Close image series.

        Stops the acquisition and sets is_acquiring flag to False.
        """
        self.cam.stop_acquisition()
        self.is_acquiring = False

    def get_new_frame(self):
        """Get frame from Ximea camera.

        Returns
        -------
        frame : numpy.ndarray
            Frame ids from Ximea camera.
        """
        # attach buffer to image object
        self._image.bp = self._data_buffer[self._frames_received].ctypes.data
        self._image.bp_size = self._data_buffer[self._frames_received].nbytes
        # get data from camera
        try:
            self.cam.get_image(self._image, 500)
        except xiapi.Xi_error as e:
            logger.error(f"Error getting image from camera: {e}")
            return []

        frames_received = [self._frames_received]

        self._frames_received += 1
        if self._frames_received >= self._number_of_frames:
            self._frames_received = 0

        return frames_received

class MU196XRCamera(XimeaBase):
    """Ximea MU196XR class.

    This camear class supports ximea camera models: MU196MR/MU196CR.
    """

    def __init__(
        self,
        microscope_name: str,
        device_connection: Any,
        configuration: Dict[str, Any],
        *args: Optional[Any],
        **kwargs: Optional[Any],
    ) -> None:
        """Initialize MU196XR class.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_connection : Any
            Hardware device to connect to
        configuration : Dict[str, Any]
            Global configuration of the microscope
        """
        super().__init__(microscope_name, device_connection, configuration)

        # on MU196 series, the trigger input is on GPI port 3, gpi index is 2
        self.cam.set_param("gpi_selector", "XI_GPI_PORT2")
        # need to reset the trigger mode to XI_GPI_TRIGGER, otherwise the trigger mode is XI_GPI_OFF
        self.cam.set_param("gpi_mode", "XI_GPI_TRIGGER")

    def __str__(str):
        """Return string representation of Ximea MU196XR class
        
        Returns
        -------
        str
            String representation of Ximea MU196XR class.
        """
        return "Ximea MU196XR Camera"

